<?php

namespace WcMipConnector\Controller;

defined('ABSPATH') || exit;

require __DIR__.'/../../vendor/autoload.php';

use WcMipConnector\Enum\MipWcConnector;
use WcMipConnector\Exception\WooCommerceApiConnectionException;
use WcMipConnector\Factory\OrderFactory;
use WcMipConnector\Manager\FileLogManager;
use WcMipConnector\Manager\OrderLogManager;
use WcMipConnector\Manager\WoocommerceApiConnectionManager;
use WcMipConnector\Service\AccountInfoReportService;
use WcMipConnector\Service\LanguageReportService;
use WcMipConnector\Service\CategoryService;
use WcMipConnector\Service\LoggerService;
use WcMipConnector\Service\OrderService;
use WcMipConnector\Service\ProductService;
use WcMipConnector\Service\StatusReportService;
use WcMipConnector\Service\SystemService;
use WcMipConnector\Enum\StatusTypes;
use WcMipConnector\Service\TaxesService;

class ApiController
{
    private const SIGN_ALGORITHM_SHA256 = 'sha256';
    private const SRC_JSON_IMPORT_FILES = '/importFiles';
    private const FILE_PRODUCT_PERMISSIONS = 0775;
    private const API_CARRIER_URL = 'http://www.bigbuy.eu/service/carriers/';

    private $apiUrl = self::API_CARRIER_URL;
    private $messageType;
    private $operationType;
    private $message;
    private $accessToken;
    private $signature;
    private $version;
    private $headers;
    private $orderLogManager;
    private $orderService;
    private $systemService;
    private $woocommerceApiConnectionManager;
    private $languageReportService;
    private $logger;

    public function __construct()
    {
        add_filter('query_vars', array($this, 'addQueryVars'), 0);
        add_action('init', array(__CLASS__, 'addEndpoints'), 0);
        add_action('parse_request', array($this, 'processAction'), 0);

        $this->orderLogManager = new OrderLogManager();
        $this->orderService = new OrderService();
        $this->systemService = new SystemService();
        $this->languageReportService = new LanguageReportService();
        $this->woocommerceApiConnectionManager = new WoocommerceApiConnectionManager();
        $loggerService = new LoggerService();
        $this->logger = $loggerService->getInstance();
    }

    public function addQueryVars($vars): array
    {
        $vars[] = 'mip-route';
        $vars[] = 'cron-route';
        return $vars;
    }

    public static function addEndpoints(): void
    {
        add_rewrite_rule('^wcmipconnector/api/(.*)?', 'index.php?mip-route=$matches[1]', 'top');
        add_rewrite_rule('^wcmipconnector/api?', 'index.php', 'top' );
        add_rewrite_rule('^wcmipconnector/(.*)?', 'index.php?cron-route=$matches[1]', 'top' );
        flush_rewrite_rules();
    }

    /**
     * @throws WooCommerceApiConnectionException
     */
    public function processAction(): void
    {
        global $wp;

        if (array_key_exists('mip-route', $wp->query_vars) && (($wp->query_vars['mip-route'] === 'authorize') || ($wp->query_vars['mip-route'] === 'grant-access'))) {
            $oauthController = new OAuthController();
            $oauthController->handleAuthRequests();
        }

        if (array_key_exists('mip-route', $wp->query_vars) && ($wp->query_vars['mip-route'] === 'order')) {
            $controller = new OrderController();
            $controller->processAction();
        }

        if (array_key_exists('cron-route', $wp->query_vars) && ($wp->query_vars['cron-route'] === 'readfiles')) {
            $fileController = new FileController();
            $fileController->executeSynchronization();
        }

        if (array_key_exists('cron-route', $wp->query_vars) && ($wp->query_vars['cron-route'] === 'updateurls')) {
            $productService = new ProductService();
            $productService->updateProductsImagesUrls();
        }

        if (!isset($_GET['messageType'], $_GET['operationType'])) {
            return;
        }

        try {
            $this->checkMethod();
            $this->headers = getallheaders();

            $this->sanitizeHeaders();
            $this->checkHeaders();

            $this->messageType = $_GET['messageType'];
            $this->operationType = $_GET['operationType'];
            $this->message = @file_get_contents('php://input');

            $this->accessToken = $this->headers['Access-Token'];
            $this->signature = $this->headers['X-Body-Signature'];
            $this->version = $this->headers['X-Version'];

            $this->checkSecurityToken();
            $this->checkSignature();

            $response = $this->executeAction();

            $this->systemService->JsonResponse(StatusTypes::HTTP_OK, 'Ok', $response);
        } catch (\Exception $e) {
            $this->systemService->JsonResponse(StatusTypes::HTTP_INTERNAL_SERVER_ERROR, 'System error');
        }
    }

    /**
     * @return array|mixed|object
     * @throws \Exception
     */
    private function executeAction()
    {
        switch ($this->messageType) {
            case 'ORDER':
                return $this->executeOrderAction();
            case 'PRODUCT':
                return $this->executeProductAction();
            case 'SYSTEM':
                return $this->executeSystemAction();
            case 'CATEGORY':
                return $this->executeCategoryAction();
            default:
                $this->systemService->JsonResponse(
                    StatusTypes::HTTP_BAD_REQUEST,
                    'The requested operation is not valid.'
                );
        }
    }

    /**
     * @return array|null
     * @throws WooCommerceApiConnectionException
     */
    private function executeOrderAction(): ?array
    {
        switch ($this->operationType) {
            case 'GET':
                return $this->listOrders();
            case 'UPDATE' :
                return $this->modifyOrders();
            case 'DELETE' :
            default:
                $this->systemService->JsonResponse(StatusTypes::HTTP_BAD_REQUEST, 'The order action is not valid.');
        }
    }

    /**
     * @return array|null
     */
    private function executeProductAction(): ?array
    {
        switch ($this->operationType) {
            case 'LIST':
            case 'UPDATE':
            case 'DELETE':
                return $this->listProduct();
            default :
                $this->systemService->JsonResponse(StatusTypes::HTTP_BAD_REQUEST, 'The product action is not valid.');
        }
    }

    /**
     * @return array
     */
    public function listOrders(): array
    {
        $orderFactory = new OrderFactory();
        $request = json_decode($this->message, true);
        $shopOrders = ['Orders' => []];

        if (empty($request['CreateTimeFrom']) && empty($request['Ids'])) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_BAD_REQUEST,
                'The CreateTimeFrom field is not defined.',
                $shopOrders
            );
        }

        if (empty($request['Ids'])) {
            $date = $request['CreateTimeFrom'];
            $ordersWithoutShipping = $this->orderLogManager->getOrdersWithoutShippingByDate($date);
        } else {
            $ordersWithoutShipping = $request['Ids'];
        }

        if (empty($ordersWithoutShipping)) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_NOT_FOUND,
                'The shop has not returned any orders.',
                $shopOrders
            );

            $this->logger->info('The shop has not returned any orders.');

            return $shopOrders;
        }

        foreach ($ordersWithoutShipping as $order) {
            if (array_key_exists('order_id', $order)) {
                $shopOrder = $this->orderService->getOrderByOrderId((int)$order['order_id']);
            } else {
                $shopOrder = $this->orderService->getOrderByOrderId((int)$order);
            }

            if ($shopOrder) {
                $shopOrders['Orders'][] = $orderFactory->create($shopOrder);
                if ($this->orderLogManager->getByOrderId((int)$shopOrder['id'])) {
                    $this->orderLogManager->setOrderAsProcess((int)$shopOrder['id']);
                    continue;
                }
                $this->orderLogManager->saveAndSetAsProcessed($shopOrder);
            }
        }

        return $shopOrders;
    }

    /**
     * @return array
     * @throws WooCommerceApiConnectionException
     */
    public function modifyOrders(): array
    {
        $request = json_decode($this->message, true);
        $responseModifyOrders = ['Orders' => []];

        foreach ($request['Orders'] as $order) {
            $orderStateIdNew = $this->orderService->stateOrderMap($order['State']);

            if (!$orderStateIdNew) {
                $this->systemService->JsonResponse(StatusTypes::HTTP_BAD_REQUEST, 'Bad Order Status', $order['State']);
            }

            $result = $this->orderService->changeStateOrder($order, $orderStateIdNew);
            $this->orderLogManager->setOrderAsUpdate((int)$order['OrderID']);
            $responseModifyOrders['Orders'][] = $result;
        }

        return $responseModifyOrders;
    }

    /**
     * @return array|null
     */
    public function listProduct(): ?array
    {
        $fileLogManager = new FileLogManager();

        $fileName = uniqid('', false);
        $fileDir = MipWcConnector::VAR_DIR.self::SRC_JSON_IMPORT_FILES;
        $fileToSave = $fileDir.'/'.$fileName.'.json';

        try {
            $oldPermissions = umask(0000);

            if (!@mkdir($fileDir) && !is_dir($fileDir)) {
                $this->systemService->JsonResponse(
                    StatusTypes::HTTP_INTERNAL_SERVER_ERROR,
                    'The file directory could not be created in '.$fileDir
                );
            }

            @chmod($fileDir, self::FILE_PRODUCT_PERMISSIONS);
            umask($oldPermissions);

            file_put_contents($fileToSave, $this->message);
        } catch (\Exception $e) {
            echo $e->getMessage();
            return null;
        }

        $fileLogManager->insert($fileName, $this->version);
        $mipFile = $fileLogManager->getByName($fileName);

        if (null === $mipFile['name']) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_INTERNAL_SERVER_ERROR,
                'File entry $fileName was not found in DB.'
            );
        }

        return ['Submission' => ['SubmissionID' => $fileName]];
    }

    private function checkMethod(): void
    {
        if ($_SERVER['REQUEST_METHOD'] !== 'POST') {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_BAD_REQUEST,
                'The method of request '.$_SERVER['REQUEST_METHOD'].' is not valid.'
            );
        }
    }

    private function sanitizeHeaders(): void
    {
        foreach ($this->headers as $header => $value) {
            $sanitizeHeader = str_replace(' ', '-', ucwords(strtolower(str_replace('-', ' ', $header))));
            $this->headers[$sanitizeHeader] = $value;
        }
    }

    private function checkHeaders(): void
    {
        if (empty($this->headers['Access-Token'])) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_BAD_REQUEST,
                'The Access-Token parameter is not defined.'
            );
        }

        if (empty($this->headers['X-Body-Signature'])) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_BAD_REQUEST,
                'The X-Body-Signature parameter is not defined.'
            );
        }

        if (empty($this->headers['X-Version'])) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_BAD_REQUEST,
                'The X-Version parameter is not defined.'
            );
        }
    }

    private function checkSecurityToken(): void
    {
        $accessToken = $this->woocommerceApiConnectionManager->getAccessToken();

        if ($this->accessToken !== $accessToken) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_BAD_REQUEST,
                'The Access-Token is not right '.$this->accessToken
            );
        }
    }

    private function checkSignature(): void
    {
        $secretKey = $this->woocommerceApiConnectionManager->getSecretKey();
        $signature = $this->sign($this->message, $secretKey);

        if ($this->signature !== $signature) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_FORBIDDEN,
                'Incorrect message signature: '.$this->signature.' // Right: '.$signature
            );
        }
    }

    private function sign(string $data, string $secretKey): string
    {
        return base64_encode(
            hash_hmac(self::SIGN_ALGORITHM_SHA256, $data, $secretKey, true)
        );
    }

    /**
     * @return array
     * @throws \Exception
     */
    private function executeSystemAction(): ?array
    {
        switch ($this->operationType) {
            case 'LANGUAGES':
                return $this->getListLanguages();
            case 'ACCOUNTINFO':
                return $this->getAccountInfoReport();
            case 'STATUSREPORT':
                return $this->getStatusReport();
            case 'TAXES':
                return $this->getTaxes();
            default :
                $this->systemService->JsonResponse(
                    StatusTypes::HTTP_BAD_REQUEST,
                    'The requested operation is not valid.'
                );
        }
    }

    /**
     * @return array
     * @throws WooCommerceApiConnectionException
     */
    private function getAccountInfoReport(): array
    {
        $accountInfoReportService = new AccountInfoReportService();

        return $accountInfoReportService->get();
    }

    /**
     * @return array
     * @throws \Exception
     */
    private function getStatusReport(): array
    {
        $statusReportService = new StatusReportService();

        return $statusReportService->get();
    }

    /**
     * @return array
     */
    private function getListLanguages(): array
    {
        $response = ['Languages' => []];
        $languages = $this->languageReportService->getDefaultLanguageIsoCode();

        if (is_string($languages)) {
            $this->systemService->JsonResponse(StatusTypes::HTTP_BAD_REQUEST, $languages, $response);
        }

        if (empty($languages)) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_NOT_FOUND,
                'The shop has not returned any language',
                $response
            );
        }

        $response = $languages;

        return $response;
    }

    /**
     * @return array|null
     */
    private function executeCategoryAction(): ?array
    {
        switch ($this->operationType) {
            case 'LIST':
            case 'UPDATE':
                return $this->listCategory();
            default :
                $this->systemService->JsonResponse(
                    StatusTypes::HTTP_BAD_REQUEST,
                    'The requested operation is not valid.'
                );
        }
    }

    /**
     * @return array
     */
    private function listCategory(): array
    {
        $request = json_decode($this->message, true);
        $categoryService = new CategoryService();

        if (is_string($request)) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_BAD_REQUEST,
                'There is an error in categories',
                array(false)
            );
            $this->logger->error('Update Categories - Wrong Request: '.$this->message);
        }

        try {
            $languages = $this->languageReportService->getShopLanguages();
            $categoryService->updateCategories($request, $languages);
            $date = new \DateTime('now');
            $this->logger->info('Update Categories - Request: '. $date->format('Y-m-d H:i:s'));
        } catch (\Exception $e) {
            $this->systemService->JsonResponse(
                StatusTypes::HTTP_BAD_REQUEST,
                'Error processing Categories',
                array(false)
            );
            $this->logger->error('Update Categories - Request: '.$this->message.' - Process Batch Message: '.$e->getMessage());
        }

        return array(true);
    }

    /**
     * @return array|string
     * @throws WooCommerceApiConnectionException
     */
    public function getTaxes()
    {
        $taxesResponse = ['Taxes' => []];
        $taxesService = new TaxesService();
        $taxes = $taxesService->getTaxes();

        if (is_string($taxes)) {
            $this->systemService->JsonResponse(StatusTypes::HTTP_BAD_REQUEST, $taxes, $taxesResponse);
        }

        if (empty($taxes)) {
            $this->systemService->JsonResponse(StatusTypes::HTTP_NOT_FOUND, 'The shop does not return any taxes', $taxesResponse);
        }

        $taxesResponse = $taxes;

        return $taxesResponse;
    }

    /**
     * @param string $apiCall
     * @param array  $params
     * @param string $language
     *
     * @return array
     */
    public function send(string $apiCall, array $params = [], string $language = 'EN'): array
    {
        $queryParams = array_merge($params, ['language' => $language]);
        $url = $this->apiUrl.$apiCall;
        $args = array(
            'timeout' => '60',
            'method' => 'GET',
            'httpversion' => '1.1',
            'body' => $queryParams
        );
        $response = wp_remote_post($url, $args);
        $responseHttpCode = wp_remote_retrieve_response_code($response);

        if ($responseHttpCode >= StatusTypes::HTTP_BAD_REQUEST && $responseHttpCode < StatusTypes::HTTP_INTERNAL_SERVER_ERROR) {
            return [
                'status' => (string)$responseHttpCode,
                'msg' => 'Client error '.$responseHttpCode.' - Failed sending '.$apiCall.' request',
            ];
        }

        if ($responseHttpCode >= StatusTypes::HTTP_INTERNAL_SERVER_ERROR && $responseHttpCode < StatusTypes::INVALID_REQUEST_CODE) {
            return [
                'status' => (string)$responseHttpCode,
                'msg' => 'Server error '.$responseHttpCode.' - API server failed to respond.',
            ];
        }

        return json_decode(wp_remote_retrieve_body($response), true);
    }
}