<?php

namespace WcMipConnector\Manager;

defined('ABSPATH') || exit;

use WcMipConnector\Model\FileReport;
use WcMipConnector\Repository\FileLogRepository;

class FileLogManager
{
    public const FILE_IN_PROCESS = 1;
    public const FILE_NOT_IN_PROCESS = 0;

    /** @var FileLogRepository */
    private $fileLogRepository;

    public function __construct()
    {
        $this->fileLogRepository = new FileLogRepository();
    }

    /**
     * @param string $fileName
     * @param string $version
     *
     * @return bool
     */
    public function insert(string $fileName, string $version): bool
    {
        $data = [
            'name' => $fileName,
            'version' => $version,
            'date_add' => date('Y-m-d H:i:s'),
        ];

        return $this->fileLogRepository->insert($data);
    }

    /**
     * @param int $idFile
     *
     * @return bool
     */
    public function update(int $idFile): bool
    {
        $data = [
            'date_process' => date('Y-m-d H:i:s'),
            'in_process' => self::FILE_NOT_IN_PROCESS
        ];

        $where = [
            'file_id' => $idFile,
        ];

        return $this->fileLogRepository->update($data, $where);
    }

    /**
     * @return \stdClass
     */
    public function getFileWithoutProcess(): ?array
    {
        return $this->fileLogRepository->getFileWithoutProcess();
    }

    /**
     * @param int $idFile
     *
     * @return bool
     */
    public function updateFileAsInProcess(int $idFile): bool
    {
        $data = [
            'in_process' => self::FILE_IN_PROCESS,
            'date_processing_start' => date('Y-m-d H:i:s')
        ];

        $where = [
            'file_id' => $idFile,
        ];

        return $this->fileLogRepository->update($data, $where);
    }

    /**
     * @return bool
     */
    public function getFileInProcess(): bool
    {
        return $this->fileLogRepository->getFileInProcess();
    }

    /**
     * @param int $day
     *
     * @return array
     */
    public function getFileLogNameByDay(int $day): array
    {
        return $this->fileLogRepository->getFileLogNameByDay($day);
    }

    /**
     * @param array $fileName
     *
     * @return bool
     */
    public function deleteFileLogByFileName(array $fileName): bool
    {
        return $this->fileLogRepository->deleteFileLogByFileName($fileName);
    }

    /**
     * @param string $fileName
     *
     * @return array|null
     */
    public function getByName(string $fileName): ?array
    {
        return $this->fileLogRepository->getByName($fileName);
    }

    /**
     * @return array|null
     */
    public function checkForLockedFile(): ?array
    {
        return $this->fileLogRepository->checkForLockedFile();
    }

    /**
     * @param int $fileId
     *
     * @return bool
     */
    public function resetFileState(int $fileId): bool
    {
        $data = [
            'in_process' => self::FILE_NOT_IN_PROCESS,
            'date_processing_start' => null,
            'date_process' => null,
        ];

        $where = [
            'file_id' => $fileId,
        ];

        return $this->fileLogRepository->update($data, $where);
    }

    /**
     * @return FileReport
     * @throws \Exception
     */
    public function getFileReport(): FileReport
    {
        $response = new FileReport();

        $response->TotalFiles = $this->fileLogRepository->getTotalFilesCount();
        $response->PendingFiles = $this->fileLogRepository->getPendingFilesCount();
        $response->ProcessingFiles = $this->fileLogRepository->getProcessingFilesCount();

        $lastStartTime = $this->fileLogRepository->getLastStartTime();
        if ($lastStartTime) {
            $timezone = new \DateTimeZone('UTC');
            $lastStartTime = new \DateTime($lastStartTime, $timezone);
            $response->LastStartTime = $lastStartTime->format(DATE_W3C);
        }else{
            $response->LastStartTime = false;
        }

        $lastAddTime =  $this->fileLogRepository->getLastAddTime();
        if ($lastAddTime) {
            $timezone = new \DateTimeZone('UTC');
            $lastAddTime = new \DateTime($lastAddTime, $timezone);
            $response->LastAddTime = $lastAddTime->format(DATE_W3C);
        }else{
            $response->LastAddTime = false;
        }

        return $response;
    }
}