<?php

namespace WcMipConnector\Manager;

defined('ABSPATH') || exit;

use WcMipConnector\Entity\ProductMap;
use WcMipConnector\Exception\WooCommerceApiConnectionException;
use WcMipConnector\Repository\ProductMapRepository;

class ProductMapManager
{
    private const PRODUCT_URL_TABLE = 'mipconnector_product_url';
    private const IMAGE_URL_TABLE = 'mipconnector_image_url';
    private const PRODUCT_IMAGE_URL_TABLE = 'mipconnector_product_image_url';

    /** @var ProductMapRepository */
    protected $productMapRepository;
    /** @var string */
    private $productUrlTable;
    /** @var string */
    private $imageUrlTable;
    /** @var string */
    private $productImageUrlTable;
    /** @var \wpdb */
    protected $woocommerceClass;

    public function __construct()
    {
        global $wpdb;
        $this->productMapRepository = new ProductMapRepository();
        $this->woocommerceClass = $wpdb;
        $this->productUrlTable = $this->woocommerceClass->prefix.self::PRODUCT_URL_TABLE;
        $this->imageUrlTable = $this->woocommerceClass->prefix.self::IMAGE_URL_TABLE;
        $this->productImageUrlTable = $this->woocommerceClass->prefix.self::PRODUCT_IMAGE_URL_TABLE;
    }

    /**
     * @param array $productMapIds
     *
     * @return array
     */
    public function findByProductMapIdsIndexedByProductMapId(array $productMapIds): array
    {
        return $this->productMapRepository->findByProductMapIdsIndexedByProductMapId($productMapIds);

    }

    /**
     * @param array $productMapIds
     *
     * @return array
     */
    public function findVersionsIndexedByProductMapId(array $productMapIds): array
    {
        return $this->productMapRepository->findVersionsIndexedByProductMapId($productMapIds);

    }

    /**
     * @param ProductMap $productMap
     *
     * @return bool
     */
    public function save(ProductMap $productMap): bool
    {
        $data = [
            'product_id' => $productMap->productId,
            'product_shop_id' => $productMap->shopProductId,
            'version' => $productMap->version,
            'date_add' => date('Y-m-d H:i:s'),
            'date_update' => null,
        ];

        return $this->productMapRepository->save($data);
    }

    /**
     * @param int $productShopId
     * @param ProductMap $productMap
     *
     * @return bool
     */
    public function update(int $productShopId, ProductMap $productMap): bool
    {
        $data = [
            'product_shop_id' => $productShopId,
            'version' => $productMap->version,
            'date_update' => date('Y-m-d H:i:s'),
        ];

        $where = [
            'product_id' => $productMap->productId,
        ];

        return $this->productMapRepository->update($data, $where);
    }

    /**
     * @return array
     */
    public function getProductsShopIndexedByProductShopId(): array
    {
        return $this->productMapRepository->getProductsShopIndexedByProductShopId();
    }

    /**
     * @param array $productsUrlIndexedByProductShopId
     * @return array
     * @throws WooCommerceApiConnectionException
     */
    public function updateProductsUrl(array $productsUrlIndexedByProductShopId): array
    {
        if (empty($productsUrlIndexedByProductShopId)) {
            return [];
        }

        $dateNow = new \DateTime('now', new \DateTimeZone('UTC'));

        $languageManager = new LanguageReportManager();
        $language = explode('_', $languageManager->getDefaultLanguageIsoCode());
        $isoCode = $language[1];

        foreach ($productsUrlIndexedByProductShopId as $productUrl) {
            $data = [
                'product_shop_id' => $productUrl['ID'],
                'iso_code' => $isoCode,
                'url' => $productUrl['guid'],
                'date_add' => $dateNow->format('Y-m-d H:i:s'),
                'date_update' => $dateNow->format('Y-m-d H:i:s'),
            ];

            $this->productMapRepository->updateProductsUrl($this->productUrlTable, $data);
        }
    }

    /**
     * @param array $productImages
     * @return array
     * @throws \Exception
     */
    public function updateImagesUrl(array $productImages): array
    {
        if (empty($productImages)) {
            return [];
        }

        $coverImages = [];
        $dateNow = new \DateTime('now', new \DateTimeZone('UTC'));

        foreach ($productImages as $productImage) {
            $cover = 0;

            if (!array_key_exists($productImage['post_parent'], $coverImages)) {
                $coverImages[$productImage['post_parent']] = $productImage['ID'];
                $cover = 1;
            }

            $data = [
                'id_image' => $productImage['ID'],
                'cover' => $cover,
                'url' => $productImage['guid'],
                'date_add' => $dateNow->format('Y-m-d H:i:s'),
                'date_update' => $dateNow->format('Y-m-d H:i:s'),
            ];

            $this->productMapRepository->updateImagesUrl($this->imageUrlTable, $data);
        }
    }

    /**
     * @param array $productImages
     * @return array
     * @throws \Exception
     */
    public function updateProductImageUrl(array $productImages): array
    {
        if (empty($productImages)) {
            return [];
        }

        $dateNow = new \DateTime('now', new \DateTimeZone('UTC'));

        foreach ($productImages as $productImage) {
            $data = [
                'product_shop_id' => $productImage['post_parent'],
                'id_image' => $productImage['ID'],
                'date_add' => $dateNow->format('Y-m-d H:i:s'),
                'date_update' => $dateNow->format('Y-m-d H:i:s'),
            ];

            $this->productMapRepository->updateProductImageUrl($this->productImageUrlTable, $data);
        }
    }

    /**
     * @return bool
     */
    public function cleanProductImagesUrls(): bool
    {
        $sqlsToTruncate[] = 'TRUNCATE TABLE '.$this->productUrlTable.';';
        $sqlsToTruncate[] = 'TRUNCATE TABLE '.$this->imageUrlTable.';';
        $sqlsToTruncate[] = 'TRUNCATE TABLE '.$this->productImageUrlTable.';';

        foreach ($sqlsToTruncate as $sql) {
            if (!$this->productMapRepository->cleanProductImagesUrls($sql)) {
                return false;
            }
        }
        return true;
    }
}