<?php

namespace WcMipConnector\Service;

defined('ABSPATH') || exit;

use WcMipConnector\Factory\AttributeFactory;
use WcMipConnector\Factory\AttributeMapFactory;
use WcMipConnector\Manager\AttributeManager;
use WcMipConnector\Manager\AttributeMapManager;
use WcMipConnector\Factory\AttributeGroupFactory;

class AttributeService
{
    /** @var AttributeManager */
    protected $attributeManager;
    /** @var AttributeMapManager */
    protected $attributeMapManager;
    /** @var AttributeFactory */
    protected $attributeFactory;
    /** @var AttributeMapFactory */
    protected $attributeMapFactory;
    /** @var AttributeGroupFactory */
    protected $attributeGroupFactory;
    /** @var AttributeGroupService */
    protected $attributeGroupService;
    /** @var ImportProcessAttributeService */
    protected $importProcessService;
    /** @var LoggerService */
    protected $logger;
    /** @var array */
    protected $languages;
    /** @var SystemService */
    protected $systemService;

    public function __construct()
    {
        $this->attributeManager = new AttributeManager();
        $this->attributeMapManager = new AttributeMapManager();
        $this->attributeFactory = new AttributeFactory();
        $this->attributeMapFactory = new AttributeMapFactory();
        $this->attributeGroupFactory = new AttributeGroupFactory();
        $this->attributeGroupService = new AttributeGroupService();
        $this->importProcessService = new ImportProcessAttributeService();
        $this->logger = new LoggerService();
        $this->systemService = new SystemService();
    }

    /**
     * @param array $attributes
     * @param int $fileId
     * @param array $languages
     * @return array
     */
    public function process(array $attributes, int $fileId, array $languages): array
    {
        $this->languages = $languages;
        $attributesGroupIndexedByAttributeGroupMapId = [];
        $attributesToProcessIndexedByParent = [];
        $attributeResult = [];

        foreach ($attributes['AttributeGroup'] as $attributeGroupData) {
            $attributeGroupFactory = $this->attributeGroupFactory->create($attributeGroupData, $languages);
            $attributesGroupIndexedByAttributeGroupMapId[$attributeGroupData['AttributeGroupID']] = $attributeGroupFactory;
        }

        foreach ($attributes['Attribute'] as $attributeParentId => $attributeData) {
            $attributesIndexedByAttributeMapId = [];
            foreach ($attributeData as $attributeChild) {
                $attributeFactory = $this->attributeFactory->create($attributeChild, $languages);
                $attributesIndexedByAttributeMapId[$attributeChild['AttributeID']] = $attributeFactory;
                $attributesToProcessIndexedByParent[$attributeParentId] = $attributesIndexedByAttributeMapId;
            }
        }


        $attributeGroup = $this->attributeGroupService->processAttributeGroupByBatch($attributesGroupIndexedByAttributeGroupMapId, $attributes['AttributeGroup'], $fileId, $languages);

        foreach ($attributesToProcessIndexedByParent as $attributeGroupParentId => $attributesToProcess){
            $attributeResult[$attributeGroupParentId] = $this->processAttributeByBatch($attributesToProcess, $attributeGroup[$attributeGroupParentId], $attributes['Attribute'][$attributeGroupParentId], $fileId);
        }

        return $attributeResult;
    }

    /**
     * @param array $attributesIndexedByAttributeMapId
     * @param int   $attributeParentId
     * @param array $attributes
     * @param int   $fileId
     *
     * @return array
     */
    public function processAttributeByBatch(array $attributesIndexedByAttributeMapId = null, int $attributeParentId = null, array $attributes = null, int $fileId = null): array
    {
        $attributeShopIndexedByAttributeId = [];
        $attributeIdsToUpdateIndexedByAttributeMapId = [];
        $processedArrays = [];
        $arrayResult = [];
        $process = [];
        $attributeShopIndexedByMapIds = [];

        if (!$attributeParentId){
            foreach ($attributesIndexedByAttributeMapId as $attributeId => $attributeData){
                $this->importProcessService->setFailure($attributeId, $fileId);
                $this->logger->getInstance()->error('AttributeID: '.$attributeId.' of FileID: '.$fileId.' does not have parent');
            }

            return [];
        }

        $batches = array_chunk($attributesIndexedByAttributeMapId, $this->systemService->getBatchValue(), true);

        foreach ($batches as $attributesToProcess) {
            $attributeFactorySave = [];
            $attributeFactoryToUpdate = [];
            $attributeToSaveIndexedByAttributeMapId = [];
            $attributeToUpdateIndexedByAttributeMapId = [];
            $tagsToNotProcess = [];

            $attributeMapsIndexedByAttributeMapId = $this->attributeMapManager->findByAttributeMapIdsIndexedByAttributeMapId(array_keys($attributesToProcess));
            $attributeVersionsIndexedByMapIds = $this->attributeMapManager->findVersionsIndexedByAttributeMapId(array_keys($attributeMapsIndexedByAttributeMapId));

            if(!empty($attributeMapsIndexedByAttributeMapId)){
                $attributeShopIndexedByAttributeId = $this->attributeManager->findByAttributeShopIdsIndexedByAttributeId(array_values($attributeMapsIndexedByAttributeMapId));
            }

            if($attributeShopIndexedByAttributeId){
                foreach ($attributeMapsIndexedByAttributeMapId as $attributeParentMappedId => $attributeShopIndexedByMapId){
                    if(array_key_exists($attributeShopIndexedByMapId, $attributeShopIndexedByAttributeId)){
                        $attributeShopIndexedByMapIds[$attributeParentMappedId] = $attributeShopIndexedByMapId;
                    }
                }
            }

            foreach ($attributesToProcess as $attributeMapId => $attributeMapData){
                if (array_key_exists($attributeMapId, $attributeShopIndexedByMapIds)){
                    $version = json_decode($attributeVersionsIndexedByMapIds[$attributeMapId], true);
                    $attributeIndexedByIsoCode = array_column($attributes[$attributeMapId]['AttributeLangs'], 'Version', 'IsoCode');
                    foreach ($version as $isoCode => $codeVersion){
                        if (array_key_exists($attributeMapId, $attributes)
                            && filter_var($codeVersion, FILTER_SANITIZE_NUMBER_INT) > $attributeIndexedByIsoCode[$isoCode]){
                            $tagsToNotProcess[$attributeMapId] = true;
                            $process[$attributeMapId] =  [
                                'id' => $attributeParentId,
                                'name' => $attributesToProcess[$attributeMapId]->name,
                            ];
                            continue;
                        }
                    }
                    if (array_key_exists($attributeMapId, $tagsToNotProcess)){
                        $processedArrays[] = $process;
                        continue;
                    }

                    $attributeMapData->id = $attributeShopIndexedByMapIds[$attributeMapId];
                    $attributeIdsToUpdateIndexedByAttributeMapId[$attributeMapId] = false;
                    $attributeFactoryToUpdate[] = $attributeMapData;
                    $attributeToUpdateIndexedByAttributeMapId[$attributeMapId] = $attributeMapData;
                    continue;
                }

                if(!array_key_exists($attributeMapId, $attributeMapsIndexedByAttributeMapId)){
                    $attributeIdsToUpdateIndexedByAttributeMapId[$attributeMapId] = false;
                    $attributeFactorySave[] = $attributeMapData;
                    $attributeToSaveIndexedByAttributeMapId[$attributeMapId] = $attributeMapData;
                    continue;
                }

                $attributeToSaveIndexedByAttributeMapId[$attributeMapId] = $attributeMapData;
                $attributeIdsToUpdateIndexedByAttributeMapId[$attributeMapId] = true;
                $attributeFactorySave[] = $attributeMapData;
            }

            $attributeRequest = ['create' => $attributeFactorySave, 'update' => $attributeFactoryToUpdate];
            $attributeBatchResponse = $this->attributeManager->save($attributeRequest, $attributeParentId);

            if (array_key_exists('create', $attributeBatchResponse) && $attributeBatchResponse['create']) {
                $processedArrays[] =  $this->save($attributeBatchResponse['create'],
                    $attributeToSaveIndexedByAttributeMapId,
                    $attributeIdsToUpdateIndexedByAttributeMapId,
                    $attributes,
                    $attributeParentId,
                    $fileId
                );
            }

            if (array_key_exists('update', $attributeBatchResponse) && $attributeBatchResponse['update']) {
                $processedArrays[] = $this->update($attributeBatchResponse['update'],
                    $attributeToUpdateIndexedByAttributeMapId,
                    $attributes,
                    $attributeParentId,
                    $fileId
                );
            }
        }

        foreach ($processedArrays as $processedArray){
            foreach ($processedArray as $processedArrayParent => $processed){
                $arrayResult[$processedArrayParent] = $processed;
            }
        }

        return $arrayResult;
    }

    /**
     * @param array $attributeBatchResponse
     * @param array  $attributeIndexedByAttributeMapId
     * @param array  $attributeIdsToUpdateIndexedByAttributeMapId
     * @param array  $attributes
     * @param int   $attributeParentId
     * @param int    $fileId
     *
     * @return array
     */
    private function save(array $attributeBatchResponse, array $attributeIndexedByAttributeMapId, array $attributeIdsToUpdateIndexedByAttributeMapId, array $attributes, int $attributeParentId, int $fileId): array
    {
        $attributeModel = [];
        $attributeMapIndex = 0;
        $attributesToDelete = null;
        foreach ($attributeIndexedByAttributeMapId as $attributeId => $attributeMapData){
            $attribute = $attributeBatchResponse[$attributeMapIndex];
            $attributeUpdated = true;
            $attributeSaved = true;

            if(!$attribute){
                continue;
            }

            if(array_key_exists('error', $attribute) && $attribute['error']){
                $importProcess = $this->importProcessService->setFailure($attributeId, $fileId);
                $this->logger->getInstance()->error('AttributeID: '.$attributeId.' in the FileID: '.$fileId.' has been not created. Batch response: '
                    .$attribute['error']['message'].'. Attribute object: '.json_encode($attributeMapData, true). 'ImportProcess mapped: '.$importProcess);
                continue;
            }

            $attributeToMap = $this->attributeMapFactory->create($attributes[$attributeId], $attribute['id'], $this->languages);

            if(array_key_exists($attributeId, $attributeIdsToUpdateIndexedByAttributeMapId) && $attributeIdsToUpdateIndexedByAttributeMapId[$attributeId]){
                $attributeUpdated = $this->attributeMapManager->update($attribute['id'], $attributeToMap);
            }else{
                $attributeSaved = $this->attributeMapManager->save($attributeToMap);
            }

            $importProcess = $this->importProcessService->setSuccess($attributeId, $fileId);
            if (!$attributeUpdated){
                $this->logger->getInstance()->error('AttributeID: '.$attributeId.' in the FileID: '.$fileId.' has been not update his map. Attribute object: '
                    .json_encode($attributeToMap, true). ' with ShopID: '.$attribute['id']. 'ImportProcess Saved: '.$importProcess);
            }

            if (!$attributeSaved){
                $attributesToDelete[$attributeId] = $attributeToMap->shopAttributeId;
                $this->logger->getInstance()->error('AttributeID: '.$attributeId.' in the FileID: '.$fileId.' has been not mapped. Attribute object: '
                    .json_encode($attributeToMap, true). '. ImportProcess Saved: '.$importProcess);
                $attributeMapIndex++;
                continue;
            }

            $attributeModel[$attributeId] =  [
                'id' => $attributeParentId,
                'name' => $attributeBatchResponse[$attributeMapIndex]['name'],
            ];

            $attributeMapIndex++;
        }

        if ($attributesToDelete){
            $attributeRequest = ['delete' => array_values($attributesToDelete)];
            $attributeBatchResponse = $this->attributeManager->save($attributeRequest, $attributeParentId);
            $this->logger->getInstance()->error('AttributeIDs: '.json_encode($attributesToDelete, true).' in the FileID: '
                .$fileId.' have been deleted. Batch Response: '.json_encode($attributeBatchResponse, true));
        }

        return $attributeModel;
    }

    /**
     * @param array $attributeBatchResponse
     * @param array $attributeIndexedByAttributeMapId
     * @param array $attributes
     * @param int   $fileId
     * @param int   $attributeParentId
     *
     * @return array
     */
    private function update(array $attributeBatchResponse, array $attributeIndexedByAttributeMapId, array $attributes, int $attributeParentId, int $fileId): array
    {
        $attributeModel = [];
        $attributeMapIndex = 0;

        foreach ($attributeIndexedByAttributeMapId as $attributeId => $attributeMapData){
            $attribute = $attributeBatchResponse[$attributeMapIndex];

            if(!$attribute){
                continue;
            }

            if(array_key_exists('error', $attribute) && $attribute['error']){
                $importProcess = $this->importProcessService->setFailure($attributeId, $fileId);
                $this->logger->getInstance()->error('AttributeID: '.$attributeId.' in the FileID: '.$fileId.' has been not created. Batch response: '
                    .$attribute['error']['message'].'. Attribute object: '.json_encode($attributeMapData, true). 'ImportProcess mapped: '.$importProcess);
                continue;
            }

            $attributeToMap = $this->attributeMapFactory->create($attributes[$attributeId], $attribute['id'], $this->languages);
            $attributeUpdated = $this->attributeMapManager->update($attribute['id'], $attributeToMap);
            $importProcess = $this->importProcessService->setSuccess($attributeId, $fileId);
            if (!$attributeUpdated){
                $this->logger->getInstance()->error('AttributeID: '.$attributeId.' in the FileID: '.$fileId.' has been not update his map. Attribute object: '
                    .json_encode($attributeToMap, true). ' with ShopID: '.$attribute['id']. 'ImportProcess Saved: '.$importProcess);
            }

            $attributeModel[$attributeId] =  [
                'id' => $attributeParentId,
                'name' => $attributeBatchResponse[$attributeMapIndex]['name'],
            ];

            $attributeMapIndex++;
        }

        return $attributeModel;
    }
}