<?php

namespace WcMipConnector\Service;

defined('ABSPATH') || exit;

use WcMipConnector\Enum\BrandTranslations;
use WcMipConnector\Exception\WooCommerceApiConnectionException;
use WcMipConnector\Factory\BrandFactory;
use WcMipConnector\Factory\BrandMapFactory;
use WcMipConnector\Manager\BrandManager;
use WcMipConnector\Manager\BrandMapManager;
use WcMipConnector\Manager\SystemManager;

class BrandService
{
    /**@var BrandFactory */
    protected $brandFactory;
    /**@var BrandMapFactory */
    protected $brandMapFactory;
    /**@var BrandMapManager */
    protected $brandMapManager;
    /**@var BrandManager */
    protected $brandManager;
    /**@var SystemManager */
    protected $systemManager;
    /**@var ImportProcessBrandService */
    protected $importProcessService;
    /**@var LoggerService */
    protected $logger;
    /** @var SystemService */
    protected $systemService;

    public function __construct()
    {
        $this->brandFactory = new BrandFactory();
        $this->brandMapFactory = new BrandMapFactory();
        $this->brandMapManager = new BrandMapManager();
        $this->brandManager = new BrandManager();
        $this->systemManager = new SystemManager();
        $this->importProcessService = new ImportProcessBrandService();
        $this->logger = new LoggerService;
        $this->systemService = new SystemService();
    }

    /**
     * @param array $brands
     * @param int $fileId
     * @param array $languages
     * @return array
     */
    public function processByBatch(array $brands, int $fileId, array $languages): array
    {
        $brandsIndexedByBrandMapId = [];
        $brandIdsToUpdateIndexedByBrandMapId = [];
        $brandsShopIndexedByBrandId = [];
        $processedArrays = [];
        $arrayResult = [];
        $process = [];
        $brandsShopIds = [];

        $brandId = $this->createBrandGroupIfNotExists($languages);
        if (!$brandId) {
            foreach ($brands as $brandId => $brandData){
                $this->importProcessService->setFailure($brandId, $fileId);
                $this->logger->getInstance()->error('BrandID: '.$brandId.' of FileID: '.$fileId.' does not have parent');
            }

            return [];
        }

        $batches = array_chunk($brands, $this->systemService->getBatchValue(), true);

        foreach ($batches as $brandsToProcess) {
            $brandFactorySave = [];
            $brandFactoryToUpdate = [];
            $brandsToSaveIndexedByBrandMapId = [];
            $brandsToUpdateIndexedByBrandMapId = [];

            foreach ($brandsToProcess as $brandData){
                $brandFactory = $this->brandFactory->create($brandData);
                $brandsIndexedByBrandMapId[$brandData['BrandID']] = $brandFactory;
            }

            $brandMapsIndexedByBrandMapId = $this->brandMapManager->findByBrandMapIdsIndexedByBrandMapId(array_keys($brandsToProcess));
            $brandVersionsIndexedByBrandMapId = $this->brandMapManager->findVersionsIndexedByBrandMapId(array_keys($brandMapsIndexedByBrandMapId));

            if(!empty($brandMapsIndexedByBrandMapId)){
                $brandsShopIds = $this->brandManager->findByBrandShopIdsIndexedByBrandId(array_values($brandMapsIndexedByBrandMapId));
            }

            if($brandsShopIds){
                foreach ($brandMapsIndexedByBrandMapId as $brandParentIdMapped => $brandIdMapped){
                    if(array_key_exists($brandIdMapped, $brandsShopIds)){
                        $brandsShopIndexedByBrandId[$brandParentIdMapped] = $brandIdMapped;
                    }
                }
            }

            foreach ($brandsIndexedByBrandMapId as $brandMapId => $brandMapData){
                if (array_key_exists($brandMapId, $brandsShopIndexedByBrandId) && (int)$brandsToProcess[$brandMapId]['Version'] < $brandVersionsIndexedByBrandMapId[$brandMapId]){
                    $process[$brandMapId] = [
                        'id'      => $brandId,
                        'visible' => true,
                        'options' => [$brandsIndexedByBrandMapId[$brandMapId]->name],
                    ];

                    $processedArrays[] = $process;
                    continue;
                }

                if(!array_key_exists($brandMapId, $brandMapsIndexedByBrandMapId)){
                    $brandIdsToUpdateIndexedByBrandMapId[$brandMapId] = false;
                    $brandFactorySave[] = $brandMapData;
                    $brandsToSaveIndexedByBrandMapId[$brandMapId] = $brandMapData;
                    continue;
                }

                if(array_key_exists($brandMapId, $brandsShopIndexedByBrandId)){
                    $brandMapData->id = $brandsShopIndexedByBrandId[$brandMapId];
                    $brandIdsToUpdateIndexedByBrandMapId[$brandMapId] = false;
                    $brandFactoryToUpdate[] = $brandMapData;
                    $brandsToUpdateIndexedByBrandMapId[$brandMapId] = $brandMapData;
                    continue;
                }

                $brandsToSaveIndexedByBrandMapId[$brandMapId] = $brandMapData;
                $brandIdsToUpdateIndexedByBrandMapId[$brandMapId] = true;
                $brandFactorySave[] = $brandMapData;
            }

            unset($brandsIndexedByBrandMapId);

            $brandRequest = ['create' => $brandFactorySave, 'update' => $brandFactoryToUpdate];

            try{
                $brandsBatchResponse = $this->brandManager->save($brandId, $brandRequest);
            } catch (WooCommerceApiConnectionException $e) {
                $this->logger->error('Brand Save - Brand Group Id: '.$brandId.' - Batch Request: '.json_encode($brandRequest, true).' - Exception Message: '.$e->getMessage());
                continue;
            }

            if (array_key_exists('create', $brandsBatchResponse) && $brandsBatchResponse['create']){
                $processedArrays[] =  $this->save($brandsBatchResponse['create'], $brandsToSaveIndexedByBrandMapId, $brandIdsToUpdateIndexedByBrandMapId, $brandsToProcess, $brandId, $fileId);
            }

            if (array_key_exists('update', $brandsBatchResponse) && $brandsBatchResponse['update']){
                $processedArrays[] = $this->update($brandsBatchResponse['update'], $brandsToUpdateIndexedByBrandMapId, $brandsToProcess, $brandId, $fileId);
            }
        }

        foreach ($processedArrays as $processedArray){
            foreach ($processedArray as $processedArrayParent => $processed){
                $arrayResult[$processedArrayParent] = $processed;
            }
        }

        return $arrayResult;
 }

    /**
     * @param array $brandsBatchResponse
     * @param array  $brandsIndexedByBrandMapId
     * @param array  $brandIdsToUpdateIndexedByBrandMapId
     * @param array  $brands
     * @param int $brandParentId
     * @param int    $fileId
     *
     * @return array
     */
    private function save(array $brandsBatchResponse, array $brandsIndexedByBrandMapId, array $brandIdsToUpdateIndexedByBrandMapId, array $brands, int $brandParentId, int $fileId): array
    {
        $brandModel = [];
        $brandMapIndex = 0;
        $brandsToDelete = null;

        foreach ($brandsIndexedByBrandMapId as $brandId => $brandMapData){
            $brand = $brandsBatchResponse[$brandMapIndex];
            $brandMapIndex++;
            $brandUpdated = true;
            $brandSaved = true;

            if(!$brand){
                continue;
            }

            if(array_key_exists('error', $brand) && $brand['error']){
                $importProcess = $this->importProcessService->setFailure((int)$brandId, $fileId);
                $this->logger->getInstance()->error('BrandID: '.$brandId.' in the FileID: '.$fileId.' has been not created. Batch response: '
                    .$brand['error']['message'].'. Brand object: '.json_encode($brandMapData, true). 'ImportProcess mapped: '.$importProcess);
                continue;
            }

            $brandToMap = $this->brandMapFactory->create($brands[$brandId], $brand['id']);

            if(array_key_exists($brandId, $brandIdsToUpdateIndexedByBrandMapId) && $brandIdsToUpdateIndexedByBrandMapId[$brandId]){
                $brandUpdated = $this->brandMapManager->update($brand['id'], $brandToMap);
            }else{
                $brandSaved = $this->brandMapManager->save($brandToMap);
            }

            $importProcess = $this->importProcessService->setSuccess((int)$brandId, $fileId);

            if (!$brandUpdated){
                $this->logger->getInstance()->error('BrandID: '.$brandId.' in the FileID: '.$fileId.' has been not update his map. Brand object: '
                    .json_encode($brandToMap, true). ' with ShopID: '.$brand['id']. 'ImportProcess Saved: '.$importProcess);
            }

            if (!$brandSaved){
                $brandsToDelete[$brandId] = $brandToMap->shopBrandId;
                $this->logger->getInstance()->error('BrandID: '.$brandId.' in the FileID: '.$fileId.' has been not mapped. Brand object: '
                    .json_encode($brandToMap, true). ' ImportProcess Saved: '.$importProcess);
                continue;
            }

            $brandModel[$brandId] =  [
                'id' => $brandParentId,
                'visible' => true,
                'options' => [$brands[$brandId]['BrandName']],
            ];
        }

        if ($brandsToDelete) {
            try {
                $brandRequest = ['delete' => array_values($brandsToDelete)];
                $this->brandManager->save($brandParentId, $brandRequest);
            } catch (WooCommerceApiConnectionException $e) {
                $this->logger->error('Brand Delete of File Id:'.$fileId.' - Brand Group Id: '.$brandParentId.' - Batch Request: '.json_encode($brandRequest, true).' - Exception Message: '.$e->getMessage());
            }
        }

        return $brandModel;
    }

    /**
     * @param array $brandsBatchResponse
     * @param array $brandsIndexedByBrandMapId
     * @param array $brands
     * @param int $brandParentId
     * @param int   $fileId
     *
     * @return array
     */
    private function update(array $brandsBatchResponse, array $brandsIndexedByBrandMapId, array $brands, int $brandParentId, int $fileId): array
    {
        $brandMapIndex = 0;
        $brandModel = [];

        foreach ($brandsIndexedByBrandMapId as $brandId => $brandMapData){
            $brand = $brandsBatchResponse[$brandMapIndex];
            $brandMapIndex++;

            if(!$brand){
                continue;
            }

            if(array_key_exists('error', $brand) && $brand['error']){
                $importProcess = $this->importProcessService->setFailure($brandId, $fileId);
                $this->logger->getInstance()->error('BrandID: '.$brandId.' in the FileID: '.$fileId.' has been not updated. Batch response: '
                    .$brand['error']['message'].'. Brand object: '.json_encode($brandMapData, true). 'ImportProcess mapped: '.$importProcess);
                continue;
            }

            $brandToMap = $this->brandMapFactory->create($brands[$brandId], $brand['id']);

            $brandUpdated = $this->brandMapManager->update($brand['id'], $brandToMap);
            $importProcess = $this->importProcessService->setSuccess((int) $brandId, $fileId);

            if (!$brandUpdated){
                $this->logger->getInstance()->error('BrandID: '.$brandId.' in the FileID: '.$fileId.' has been not update his map. Brand object: '
                    .json_encode($brandToMap, true). ' with ShopID: '.$brand['id']. 'ImportProcess Saved: '.$importProcess);
            }

            $brandModel[$brandId] =  [
                'id' => $brandParentId,
                'visible' => true,
                'options' => [$brands[$brandId]['BrandName']],
            ];
        }

        return $brandModel;
    }

    /**
     * @param array $languages
     * @return int|null
     */
    private function createBrandGroupIfNotExists(array $languages): ?int
    {
        $brandId = $this->systemManager->getBrandAttributeId();

        if ($brandId) {
            $brandShopId = $this->brandManager->findById($brandId);

            if ($brandShopId) {
                return $brandId;
            }
        }

        foreach ($languages as $isoCode) {
            $brandName = BrandTranslations::getBrandTranslation(strtoupper($isoCode));
        }
        $brandGroup = $this->brandFactory->createGroup($brandName);

        try {
            $brand = $this->brandManager->create($brandGroup);
        } catch (WooCommerceApiConnectionException $e) {
            $this->logger->getInstance()->error('Create Brand Group - Exception message: '.$e->getMessage().
                ' - BrandParent Object: '.json_encode($brandGroup, true));
            return null;
        }

        $this->systemManager->updateBrandAttributeId($brand['id']);

        return $brand['id'];
    }
}