<?php

namespace WcMipConnector\Service;

defined('ABSPATH') || exit;

use WcMipConnector\Enum\MipWcConnector;
use WcMipConnector\Manager\FileLogManager;
use WcMipConnector\Manager\ImportProcessProductManager;

class FileService
{
    public const IMPORT_FILES_FOLDER = '/importFiles/';

    /** @var FileLogManager */
    private $fileLogManager;
    /** @var ImportProcessProductManager */
    private $importProcessProductManager;
    /** @var BrandService */
    private $brandService;
    /** @var TagService  */
    private $tagService;
    /** @var CategoryService  */
    private $categoryService;
    /** @var AttributeService  */
    private $attributeService;
    /** @var ProductService  */
    private $productService;
    /** @var LoggerService */
    protected $logger;
    /** @var UnPublishProductService */
    protected $disableService;

    public function __construct()
    {
        $this->fileLogManager = new FileLogManager();
        $this->importProcessProductManager = new ImportProcessProductManager();
        $this->brandService = new BrandService();
        $this->tagService = new TagService();
        $this->categoryService = new CategoryService();
        $this->attributeService = new AttributeService();
        $this->productService = new ProductService();
        $this->logger = new LoggerService;
        $this->disableService = new UnPublishProductService;
    }

    /**
     * @return void
     * @throws \Exception
     */
    public function processFile(): void
    {
        $systemService = new SystemService();
        $fileInProcess = $this->fileLogManager->getFileInProcess();
        if ($fileInProcess) {
            $this->logger->getInstance()->info('Cannot synchronize a new file because another file is in process');
            return;
        }

        $fileWithoutProcess = $this->fileLogManager->getFileWithoutProcess();

        if (!isset($fileWithoutProcess['file_id'], $fileWithoutProcess['version']) || (isset($fileWithoutProcess['file_id']) && !$fileWithoutProcess['file_id'])) {
            $this->logger->getInstance()->info('There not exists any file to be processed');
            return;
        }

        if (!$systemService->checkIfPluginNeedUpgrade($fileWithoutProcess['version'])) {
            return;
        }

        $productsArray = $this->getProductsFromFile($fileWithoutProcess);

        if (empty($productsArray)) {
            $this->fileLogManager->updateFileAsInProcess($fileWithoutProcess['file_id']);
            $this->fileLogManager->update($fileWithoutProcess['file_id']);
            $this->logger->getInstance()->warning('FileId: '.$fileWithoutProcess['file_id'].' synchronization ended because the product array is empty');
            return;
        }

        $this->fileLogManager->updateFileAsInProcess($fileWithoutProcess['file_id']);
        $productsArrayIndexedByProductId = array_column($productsArray, 'ProductID');
        $processedProducts = $this->importProcessProductManager->getProcessedProducts($productsArrayIndexedByProductId, $fileWithoutProcess['file_id']);

        $this->processProduct($productsArray, $processedProducts, $fileWithoutProcess);
        $this->fileLogManager->update($fileWithoutProcess['file_id']);

    }

    /**
     * @param array $productsArray
     * @param array $processedProducts
     * @param array $fileWithoutProcess
     *
     * @throws \Exception
     */
    private function processProduct(array $productsArray, array $processedProducts, array $fileWithoutProcess): void
    {
        $tags = [];
        $categories = [];
        $brands = [];
        $products = [];
        $tagIds = [];
        $categoriesIds = [];
        $brandsIds = [];
        $attributesIds = [];
        $attributeGroup = [];
        $attribute = [];
        $attributes = [];
        $productsDisabled = [];
        $product = [];

        foreach ($productsArray as $product) {
            if (in_array($product['ProductID'], $processedProducts, true)) {
                $this->logger->getInstance()->info('The product '.$product['ProductID'].' has already been processed successfully in the fileID '.$fileWithoutProcess['file_id']);
                continue;
            }

            if($product['Active'] === 0){
                $productsDisabled[$product['ProductID']] = $product;
                continue;
            }

            if ($product['Tags']){
                $tags = $this->getTagsFromFile($product['Tags'], $tags);
            }

            if ($product['Categories']){
                $categories = $this->getCategoriesFromFile($product['Categories'], $categories);
            }

            if (isset($product['Brand']['BrandID'])) {
                $brands = $this->getBrandsFromFile($product['Brand'], $brands);
            }

            if ($product['Variations']){
                $attributesResult = $this->getAttributesFromVariationData($product['Variations']);
                if ($attributesResult['AttributeGroup']){
                    foreach ($attributesResult['AttributeGroup'] as $attributesParent => $attributeResult){
                        $attributeGroup[$attributesParent] = $attributeResult;
                    }
                }

                if ($attributesResult['Attribute']){
                    foreach ($attributesResult['Attribute'] as $attributeParent => $attributeResult){
                        foreach ($attributeResult as $parent => $result) {
                            $attribute[$attributeParent][$parent] = $result;
                        }
                    }
                }
            }

            $products[] = $this->getProductsAndVariationsFromFile($product);
        }

        if ($attributeGroup && $attribute){
            $attributes = ['AttributeGroup' => $attributeGroup, 'Attribute' => $attribute];
        }

        if ($tags) {
          $tagIds = $this->tagService->processByBatch($tags, $fileWithoutProcess['file_id'], $product['Languages']);
        }

        if ($categories) {
          $categoriesIds = $this->categoryService->processByBatch($categories, $fileWithoutProcess['file_id'], $product['Languages']);
        }

        if ($brands) {
          $brandsIds = $this->brandService->processByBatch($brands, $fileWithoutProcess['file_id'], $product['Languages']);
        }

        if ($attributes) {
          $attributesIds = $this->attributeService->process($attributes, $fileWithoutProcess['file_id'], $product['Languages']);
        }

        if ($products){
            $this->productService->process($products, $tagIds, $categoriesIds, $brandsIds, $attributesIds, $fileWithoutProcess['file_id'], $product['Languages']);
        }

        if ($productsDisabled){
            $this->disableService->disableByBatch($productsDisabled, $fileWithoutProcess['file_id']);
        }
    }

    /**
     * @param array $productTags
     * @param array $tags
     *
     * @return array
     */
    private function getTagsFromFile(array $productTags, array $tags): array
    {
        $tagActive = get_option('WC_MIPCONNECTOR_TAG_ACTIVE');
        $tagName = get_option('WC_MIPCONNECTOR_TAG_NAME');
        foreach ($productTags as $productTag){
            if (array_key_exists('TagName', $productTag) && !$tagActive && $productTag['TagName'] === $tagName) {
                continue;
            }

            $tags[$productTag['TagID']] = $productTag;
        }

        return $tags;
    }

    /**
     * @param array $productCategories
     * @param array $categories
     *
     * @return array
     */
    private function getCategoriesFromFile(array $productCategories, array $categories): array
    {
        foreach ($productCategories as $productCategory){
            $categories[$productCategory['CategoryID']] = $productCategory;
        }

        return $categories;
    }

    /**
     * @param array $productBrands
     * @param array $brands
     *
     * @return array
     */
    private function getBrandsFromFile(array $productBrands, array $brands): array
    {
        $brands[$productBrands['BrandID']] = $productBrands;

        return $brands;
    }

    /**
     * @param array $productVariations
     *
     * @return array
     */
    private function getAttributesFromVariationData(array $productVariations): array
    {
        $attributeGroupList = [];
        $attributeList = [];

        foreach ($productVariations as $productVariation){
            foreach ($productVariation['VariationAttributes'] as $productAttribute){
                $attributeGroupList[$productAttribute['AttributeGroup']['AttributeGroupID']] = $productAttribute['AttributeGroup'];
                $attributeList[$productAttribute['AttributeGroup']['AttributeGroupID']][$productAttribute['Attribute']['AttributeID']] = $productAttribute['Attribute'];
            }
        }

        return ['AttributeGroup' => $attributeGroupList, 'Attribute' => $attributeList];
    }

    /**
     * @param array $product
     *
     * @return array
     */
    private function getProductsAndVariationsFromFile(array $product): array
    {
        $productList = [];
        $variationList = [];

        $productList[$product['ProductID']] = $product;

        if (!isset($product['Variations'])){
            $productArray = ['Product' => $productList];
            $variationArray = ['Variation' => $variationList];

            return array_merge($productArray, $variationArray);
        }

        foreach ($product['Variations'] as $variationData){
            $variationList[$product['ProductID']][$variationData['VariationID']] = $variationData;
        }

        $productArray = ['Product' => $productList];
        $variationArray = ['Variation' => $variationList];

        return array_merge($productArray, $variationArray);
    }

    /**
     * @param array
     *
     * @return array
     */
    public function getProductsFromFile(array $infoFile): array
    {
        try {
            $filename = MipWcConnector::VAR_DIR.self::IMPORT_FILES_FOLDER.$infoFile['name'].'.json';

            if (!file_exists($filename)) {
                $this->logger->getInstance()->error('The file '.$infoFile['name'].'.json does not exist in the directory '.MipWcConnector::VAR_DIR.self::IMPORT_FILES_FOLDER);
                return [];
            }

            $data = json_decode(file_get_contents($filename), true);

            if (!$data || empty($data['Products'])) {
                $this->logger->getInstance()->warning('The file '.$infoFile['name'].' doesnt contain a valid array or doesnt contain ProductData');
                return [];
            }

            return $data['Products'];
        } catch (\Exception $exception) {
            $this->logger->getInstance()->error('Error in the function getProductsFromFile(), Exception: '.$exception);
            return [];
        }
    }
}