<?php

namespace WcMipConnector\Service;

defined('ABSPATH') || exit;

use WcMipConnector\Exception\WooCommerceApiConnectionException;
use WcMipConnector\Factory\OrderFactory;
use WcMipConnector\Manager\OrderManager;
use WcMipConnector\Enum\WooCommerceOrderStatus;

class OrderService
{
    private const ORDER_STATUS_IN_PROCESS = 'ORDER_IN_PROCESS';
    private const ORDER_STATUS_SENT = 'ORDER_SENT';
    private const ORDER_STATUS_PENDING_REVIEW = 'ORDER_PENDING_REVIEW';
    private const ORDER_STATUS_DELIVERED = 'ORDER_DELIVERED';
    private const ORDER_STATUS_CANCELLED = 'ORDER_CANCELLED';

    private $orderManager;
    private $logger;

    public function __construct()
    {
        $this->orderManager = new OrderManager();
        $loggerService = new LoggerService();
        $this->logger = $loggerService->getInstance();
    }

    /**
     * @param int $orderId
     * @return array|null
     */
    public function getOrderByOrderId(int $orderId): ?array
    {
        try {
            $order = $this->orderManager->getOrderByOrderId($orderId);
        } catch (\Exception $e) {
            $this->logger->error('Getting Order Id: '.$orderId .' - Error message: '. $e->getMessage());
            return null;
        }
        return $order;
    }

    /**
     * @param string $stateOrder
     * @return string|null
     */
    public function stateOrderMap(string $stateOrder): ?string
    {
        switch ($stateOrder) {
            case self::ORDER_STATUS_CANCELLED:
                return WooCommerceOrderStatus::CANCELLED;
            case self::ORDER_STATUS_DELIVERED:
                return WooCommerceOrderStatus::COMPLETED;
            case self::ORDER_STATUS_SENT:
            case self::ORDER_STATUS_IN_PROCESS:
                return WooCommerceOrderStatus::PROCESSING;
            case self::ORDER_STATUS_PENDING_REVIEW:
                return WooCommerceOrderStatus::ON_HOLD;
            default:
                $this->logger->error('Order Status not valid: ' . $stateOrder);
                return null;
        }
    }

    /**
     * @param array $orderData
     * @param string $orderNewStatus
     * @return array
     */
    public function changeStateOrder(array $orderData, string $orderNewStatus): array
    {
        $orderFactory = new OrderFactory();
        $response = [
            'OrderID' => $orderData['OrderID'],
            'State' => false,
        ];

        $order = $this->getOrderByOrderId((int)$orderData['OrderID']);

        if (!$order || (((string)$order['status'] === $orderNewStatus) && $orderData['State'] !== self::ORDER_STATUS_SENT)) {
            $this->logger->error('Order status not updated in order id: '.$orderData['OrderID'].' - Current status: '.$order['status']. ' - New status: '.$orderNewStatus.' - Order data request: '.json_encode($orderData));
            return $response;
        }

        if($orderData['TrackingNumber']){
            $orderNote = $orderFactory->setOrderTracking($orderData);
            try {
                $this->orderManager->addOrderNote($order, $orderNote);
            } catch (WooCommerceApiConnectionException $exception) {
                $this->logger->error('Adding tracking in order id: '.$order['id'] .' - Order data request: '.json_encode($orderData).' - Error message: '. $exception->getMessage());
            }
        }

        try {
            $order = $orderFactory->setOrderStatus($order, $orderNewStatus);
            $this->orderManager->update($order);
            $response['State'] = true;
        } catch (\Exception $e) {
            $this->logger->error('Changing order status in Order id: '.$order['id'] .' - New status: '.$orderNewStatus.' - Order data request: '.json_encode($orderData).' - Error message: '. $e->getMessage());
            $response['State'] = false;
        }
        return $response;
    }
}