<?php

namespace WcMipConnector\Service;

defined('ABSPATH') || exit;

use WcMipConnector\Exception\WooCommerceApiConnectionException;
use WcMipConnector\Factory\ProductFactory;
use WcMipConnector\Factory\ProductMapFactory;
use WcMipConnector\Manager\BrandManager;
use WcMipConnector\Manager\ProductManager;
use WcMipConnector\Manager\ProductMapManager;
use WcMipConnector\Factory\VariationFactory;

class ProductService
{
    /** @var ProductManager */
    protected $productManager;
    /** @var ProductMapManager */
    protected $productMapManager;
    /** @var ProductFactory */
    protected $productFactory;
    /** @var ProductMapFactory */
    protected $productMapFactory;
    /** @var VariationFactory */
    protected $variationFactory;
    /** @var VariationService */
    protected $variationService;
    /** @var ImportProcessProductService */
    protected $importProcessService;
    /** @var BrandManager */
    protected $brandManager;
    /** @var LoggerService */
    protected $logger;
    /** @var array */
    protected $languages;
    /** @var SystemService */
    protected $systemService;

    public function __construct()
    {
        $this->productManager = new ProductManager();
        $this->productMapManager = new ProductMapManager();
        $this->productFactory = new ProductFactory();
        $this->productMapFactory = new ProductMapFactory();
        $this->variationFactory = new VariationFactory();
        $this->variationService = new VariationService();
        $this->importProcessService = new ImportProcessProductService();
        $this->brandManager = new BrandManager();
        $this->logger = new LoggerService;
        $this->systemService = new SystemService();
    }

    /**
     * @param array|null $products
     * @param array|null $tags
     * @param array|null $categories
     * @param array|null $brands
     * @param array|null $attributes
     * @param int|null $fileId
     * @param array $languages
     * @throws \Exception
     */
    public function process(array $products = null, array $tags = null, array $categories = null, array $brands = null, array $attributes = null, int $fileId = null, array $languages = null): void
    {
        $this->languages = $languages;
        $productIndexedByProductMapId = [];
        $variationsToProcessIndexedByParent = [];
        $productToProcess = [];
        $variationToProcess = [];

        foreach ($products as $product) {
            foreach ($product['Product'] as $productId => $productData) {
                $productAttributes = [];
                $tagIds = $this->getTagIdsByProductTags($productData['Tags'], $tags);
                $categoryIds = $this->getCategoryIdsByProductCategories($productData['Categories'], $categories);
                $brandIds = $this->getBrandIdsByProductBrands($productData['Brand'], $brands);

                if ($productData['Variations']){
                    $productAttributes = $this->getAttributesByVariations($productData['Variations'], $attributes);
                }

                $productToProcess[$productId] = $productData;
                $productFactory = $this->productFactory->create($productData, $tagIds, $categoryIds, $brandIds, $productAttributes, $languages);

                $productIndexedByProductMapId[$productData['ProductID']] = $productFactory;
            }

            if (!$product['Variation']){
                continue;
            }

            foreach ($product['Variation'] as $productParentId => $variationData) {
                $variationsIndexedByVariationMapId = [];
                foreach ($variationData as $variationChild) {
                    $attributesIds = $this->getAttributesIdsByVariations($variationChild['VariationAttributes'], $attributes);

                    $variationFactory = $this->variationFactory->create($variationChild, $attributesIds);

                    $variationToProcess[$productParentId][$variationChild['VariationID']] = $variationChild;
                    $variationsIndexedByVariationMapId[$variationChild['VariationID']] = $variationFactory;
                    $variationsToProcessIndexedByParent[$productParentId] = $variationsIndexedByVariationMapId;
                }
            }
        }

        $productParentProcessed = $this->processProductByBatch($productIndexedByProductMapId, $productToProcess, $fileId);

        foreach ($variationsToProcessIndexedByParent as $variationParentId => $variationsToProcess){
            $this->variationService->processVariationByBatch($variationsToProcess, $productParentProcessed[$variationParentId], $variationToProcess[$variationParentId], $fileId);
        }
    }

    /**
     * @param array $productTags
     * @param array $tags
     *
     * @return array
     */
    private function getTagIdsByProductTags(array $productTags, array $tags): array
    {
        $tagIds = [];

        foreach ($productTags as $productTag){
            if(array_key_exists($productTag['TagID'], $tags)){
                $tagIds[] = ['id' => $tags[$productTag['TagID']]];
            }
        }

        return $tagIds;
    }

    /**
     * @param array $productCategories
     * @param array $categories
     *
     * @return array
     */
    private function getCategoryIdsByProductCategories(array $productCategories, array $categories): array
    {
        $categoryIds = [];

        foreach ($productCategories as $productCategory){
            if(array_key_exists($productCategory['CategoryID'], $categories)){
                $categoryIds[] = ['id' => $categories[$productCategory['CategoryID']]];
            }
        }

        return $categoryIds;
    }

    /**
     * @param array $productBrand
     * @param array $brandsData
     *
     * @return array
     */
    private function getBrandIdsByProductBrands(array $productBrand, array $brandsData): array
    {
        $brandIds = [];

        if(array_key_exists($productBrand['BrandID'], $brandsData)){
            return $brandsData[$productBrand['BrandID']];
        }

        return $brandIds;
    }

    /**
     * @param array $variations
     * @param array $attributes
     *
     * @return array
     */
    private function getAttributesIdsByVariations(array $variations, array $attributes): array
    {
        $attributeIds = [];

        foreach ($variations as $variation){

            $attributeGroupId = $variation['AttributeGroup']['AttributeGroupID'];
            $attributeId = $variation['Attribute']['AttributeID'];

            if(array_key_exists($attributeGroupId, $attributes) && array_key_exists($attributeId, $attributes[$attributeGroupId])){
                $attributeIds[] = $attributes[$attributeGroupId][$attributeId];
            }
        }

        return $attributeIds;
    }

    /**
     * @param array $variations
     * @param array $attributesIndexedByAttributeGroupId
     *
     * @return array
     */
    private function getAttributesByVariations(array $variations, array $attributesIndexedByAttributeGroupId): array
    {
        $attributes  = [];
        $attributeGroupOptions = [];
        $attributeIdIndexedByAttributeGroupId = [];

        foreach ($variations as $variation){
            foreach ($variation['VariationAttributes'] as $variationAttributes){
                $attributeGroupId = $variationAttributes['AttributeGroup']['AttributeGroupID'];
                $attributeId = $variationAttributes['Attribute']['AttributeID'];

                if(array_key_exists($attributeGroupId, $attributesIndexedByAttributeGroupId) && array_key_exists($attributeId, $attributesIndexedByAttributeGroupId[$attributeGroupId])){
                    $attributeGroupOptions[$attributeGroupId][] = $attributesIndexedByAttributeGroupId[$attributeGroupId][$attributeId]['name'];
                    $attributeIdIndexedByAttributeGroupId[$attributeGroupId] = $attributeId;
                }
            }
        }

        foreach ($attributeGroupOptions as $attributeGroupId => $optionNames){
            $attributes[] = [
                'id' => $attributesIndexedByAttributeGroupId[$attributeGroupId][$attributeIdIndexedByAttributeGroupId[$attributeGroupId]]['id'],
                'visible' => true,
                'variation' => true,
                'options' => $optionNames,
            ];
        }

        return $attributes;
    }

    /**
     * @param array $productsIndexedByProductMapId
     * @param array $products
     * @param int   $fileId
     *
     * @return array
     * @throws \Exception
     */
    public function processProductByBatch(array $productsIndexedByProductMapId, array $products, int $fileId): array
    {
        $processedArray = [[]];
        $productShopIds = [];
        $productShopIdsIndexedByProductMapId = [];

        if (!$productsIndexedByProductMapId){
            $productsIndexedByProductMapId = [];
        }

        $batches = array_chunk($productsIndexedByProductMapId, $this->systemService->getBatchValue(), true);
        foreach ($batches as $batchToProcess) {
            $productFactorySave = [];
            $productFactoryToUpdate = [];
            $productIdsToUpdateIndexedByProductMapId = [];
            $productToSaveIndexedByProductMapId = [];
            $productToUpdateIndexedByProductMapId = [];
            $productNotMappedIndexedById = [];
            $overrideTitleIndexedByProductId = [];
            $overrideDescriptionIndexedByProductId = [];

            $productMapIdsIndexedByProductMapId = $this->productMapManager->findByProductMapIdsIndexedByProductMapId(array_keys($batchToProcess));
            $productVersionsIndexedById = $this->productMapManager->findVersionsIndexedByProductMapId(array_keys($productMapIdsIndexedByProductMapId));
            $productsNotMappedIndexedBySku = $this->productManager->findProductsNotMappedIndexedBySku(array_keys(array_column($products, 'ProductID', 'SKU')));

            if(!empty($productMapIdsIndexedByProductMapId)){
                $productShopIds = $this->productManager->findByProductShopIdsIndexedByProductId(array_values($productMapIdsIndexedByProductMapId));
            }

            if($productShopIds){
                foreach ($productMapIdsIndexedByProductMapId as $productParentIdMapped => $productIdMapped){
                    if(array_key_exists($productIdMapped, $productShopIds)){
                        $productShopIdsIndexedByProductMapId[$productParentIdMapped] = $productIdMapped;
                    }
                }
            }

            foreach ($batchToProcess as $productMapId => $productMapData) {
                if (array_key_exists($productMapId, $productShopIdsIndexedByProductMapId)) {
                    $version = json_decode($productVersionsIndexedById[$productMapId], true);

                    if($products[$productMapId]['Version'] < $version['version']){
                        $processedArray[][] = ['shopId' => (int)$productShopIdsIndexedByProductMapId[$productMapId], 'id' => (int)$productMapId];
                        continue;
                    }

                    $productMapData->id = $productShopIdsIndexedByProductMapId[$productMapId];
                    $productIdsToUpdateIndexedByProductMapId[$productMapId] = false;

                    foreach ($products[$productMapId]['ProductLangs'] as $key => $productLang) {
                        if (array_key_exists($productLang['IsoCode'], $version) && $version[$productLang['IsoCode']] > $productLang['Version']) {
                            unset($productMapData->description, $productMapData->name);
                            $products[$productMapId]['ProductLangs'][$key]['Version'] = $version[$productLang['IsoCode']];
                            continue;
                        }

                        if (array_key_exists($productLang['IsoCode'], $version)) {
                            $overrideTitleIndexedByProductId[$productMapId] = $productLang['OverrideTitle'];
                            $overrideDescriptionIndexedByProductId[$productMapId] = $productLang['OverrideDescription'];
                            continue;
                        }
                    }

                    if(array_key_exists($productMapId, $overrideTitleIndexedByProductId) && !$overrideTitleIndexedByProductId[$productMapId]) {
                        unset($productMapData->name);
                    }

                    if(array_key_exists($productMapId, $overrideDescriptionIndexedByProductId) && !$overrideDescriptionIndexedByProductId[$productMapId]) {
                        unset($productMapData->description);
                    }

                    $productFactoryToUpdate[] = $productMapData;
                    $productToUpdateIndexedByProductMapId[$productMapId] = $productMapData;
                    continue;
                }

                if(!array_key_exists($productMapId, $productMapIdsIndexedByProductMapId)){
                    if(array_key_exists($productMapData->sku, $productsNotMappedIndexedBySku)){
                        $productMapData->id = $productsNotMappedIndexedBySku[$productMapData->sku];
                        $productFactoryToUpdate[] = $productMapData;
                        $productToUpdateIndexedByProductMapId[$productMapId] = $productMapData;
                        $productNotMappedIndexedById[$productMapId] = $productMapId;
                        continue;
                    }

                    $productIdsToUpdateIndexedByProductMapId[$productMapId] = false;
                    $productFactorySave[] = $productMapData;
                    $productToSaveIndexedByProductMapId[$productMapId] = $productMapData;
                    continue;
                }

                if (!array_key_exists($productMapId, $productShopIdsIndexedByProductMapId)) {
                    if(array_key_exists($productMapData->sku, $productsNotMappedIndexedBySku)){
                        $productMapData->id = $productsNotMappedIndexedBySku[$productMapData->sku];
                        $productFactoryToUpdate[] = $productMapData;
                        $productToUpdateIndexedByProductMapId[$productMapId] = $productMapData;
                        continue;
                    }
                    $productIdsToUpdateIndexedByProductMapId[$productMapId] = true;
                    $productFactorySave[] = $productMapData;
                    $productToSaveIndexedByProductMapId[$productMapId] = $productMapData;
                    continue;
                }
            }

            $productRequest = ['create' => $productFactorySave, 'update' => $productFactoryToUpdate];

            try {
                $productBatchResponse = $this->productManager->save($productRequest);
            } catch (WooCommerceApiConnectionException $exception) {
                $this->logger->error('Product Save - Exception Message: '.$exception->getMessage());
                continue;
            }

            if (array_key_exists('create', $productBatchResponse) && $productBatchResponse['create']) {
                $processedArray[] = $this->save($productBatchResponse['create'], $productToSaveIndexedByProductMapId, $productIdsToUpdateIndexedByProductMapId, $products, $fileId);
            }

            if (array_key_exists('update', $productBatchResponse) && $productBatchResponse['update']) {
                $processedArray[] = $this->update($productBatchResponse['update'], $productToUpdateIndexedByProductMapId, $productNotMappedIndexedById, $products, $fileId);
            }
        }
        $processed = array_merge(...$processedArray);

        return array_column($processed, 'shopId', 'id');
    }

    /**
     * @param array $productBatchResponse
     * @param array $productIndexedByProductMapId
     * @param array $productIdsToUpdateIndexedByProductMapId
     * @param array $products
     * @param int   $fileId
     *
     * @return array
     * @throws \Exception
     */
    private function save(array $productBatchResponse, array $productIndexedByProductMapId, array $productIdsToUpdateIndexedByProductMapId, array $products, int $fileId): array
    {
        $productIds = [];
        $productMapIndex = 0;
        $productUpdated = true;
        $productSaved = true;
        $productsToDelete = null;

        foreach ($productIndexedByProductMapId as $productId => $productMapData){
            $product = $productBatchResponse[$productMapIndex];
            $productMapIndex++;

            if(!$product){
                continue;
            }

            if(array_key_exists('error', $product) && $product['error']){
                $importProcess = $this->importProcessService->setFailure($productId, $fileId);
                $this->logger->getInstance()->error('Product '.$productId.' of fileID '.$fileId.' has been not created. Batch response: '
                    . $product['error']['message'].' - Product SKU: '.$productMapData->sku.' - ImportProcess added: '.$importProcess);
                continue;
            }

            $productToMap = $this->productMapFactory->create($products[$productId], $product['id'], $this->languages);

            if (array_key_exists($productId, $productIdsToUpdateIndexedByProductMapId) && $productIdsToUpdateIndexedByProductMapId[$productId]) {
                $productUpdated = $this->productMapManager->update($product['id'], $productToMap);
            } else {
                $productSaved = $this->productMapManager->save($productToMap);
            }

            $importProcess = $this->importProcessService->setSuccess($productId, $fileId);
            if(!$productUpdated){
                $productsToDelete[$productId] = $product['id'];
                $this->logger->getInstance()->error('Product '.$productId.' of fileID '.$fileId.' has been not updated in his map - Product SKU: '.$productMapData->sku.' with shopID: '.$product['id'].' - ImportProcess added: '.$importProcess);
                continue;
            }

            if(!$productSaved){
                $productsToDelete[$productId] = $productToMap->shopProductId;
                $this->logger->getInstance()->error('Product '.$productId.' of fileID '.$fileId.' has been not mapped - Product SKU: '.$productMapData->sku.' - ImportProcess added: '.$importProcess);
                continue;
            }

            $productIds[] = ['shopId' => $product['id'], 'id' => $productId];
        }

        if ($productsToDelete){
            $productsRequest = ['delete' => array_values($productsToDelete)];

            try {
                $this->productManager->save($productsRequest);
            } catch (WooCommerceApiConnectionException $exception) {
                $this->logger->error('Product Delete - Exception Message: '.$exception->getMessage());
            }

            $this->logger->getInstance()->info('ProductIDs: '.json_encode(array_keys($productsToDelete), true).' in the FileID: '
                .$fileId.' have been deleted.');
        }

        return $productIds;
    }

    /**
     * @param array $productBatchResponse
     * @param array $productIndexedByProductMapId
     * @param array $productNotMappedIndexedById
     * @param array $products
     * @param int   $fileId
     *
     * @return array
     * @throws \Exception
     */
    private function update(array $productBatchResponse, array $productIndexedByProductMapId, array $productNotMappedIndexedById, array $products, int $fileId): array
    {
        $productIds = [];
        $productMapIndex = 0;
        foreach ($productIndexedByProductMapId as $productId => $productMapData){
            $product = $productBatchResponse[$productMapIndex];
            $productMapIndex++;

            if(!$product){
                continue;
            }

            if(array_key_exists('error', $product) && $product['error']){
                $importProcess = $this->importProcessService->setFailure($productId, $fileId);
                $this->logger->getInstance()->error('Product '.$productId.' of fileID '.$fileId.' has been not updated. Batch response: '
                    .$product['error']['message'].' - Product SKU: '.$productMapData->sku.' - ImportProcess added: '.$importProcess);
                continue;
            }

            $productToMap = $this->productMapFactory->create($products[$productId], $product['id'], $this->languages);

            if(array_key_exists($productId, $productNotMappedIndexedById)){
                $productMapped = $this->productMapManager->save($productToMap);
            }
            else {
                $productMapped = $this->productMapManager->update($product['id'], $productToMap);
            }


            $importProcess = $this->importProcessService->setSuccess($productId, $fileId);
            if(!$productMapped){
                $this->logger->getInstance()->warning('Product '.$productId.' of fileID '.$fileId.' has been not updated his map - Product SKU: '.$productMapData->sku.' with shopID: '.$product['id'].' - ImportProcess added: '.$importProcess);
            }

            $productIds[] = ['shopId' => $product['id'], 'id' => $productId];
        }

        return $productIds;
    }

    /**
     * @return bool
     * @throws WooCommerceApiConnectionException
     * @throws \Exception
     */
    public function updateProductsImagesUrls(): bool
    {
        $cleanProductImagesUrls = $this->productMapManager->cleanProductImagesUrls();

        if (!$cleanProductImagesUrls) {
            return false;
        }

        $mappedProductsIndexedByProductShopId = $this->productMapManager->getProductsShopIndexedByProductShopId();
        $productsUrlIndexedByProductShopId = $this->productManager->getProductUrls($mappedProductsIndexedByProductShopId);
        $productsImage = $this->productManager->getProductImages($mappedProductsIndexedByProductShopId);

        $this->productMapManager->updateProductsUrl($productsUrlIndexedByProductShopId);
        $this->productMapManager->updateImagesUrl($productsImage);
        $this->productMapManager->updateProductImageUrl($productsImage);

        return true;
    }
}