<?php

namespace WcMipConnector\Service;

use WcMipConnector\Controller\ApiController;
use WcMipConnector\Exception\WooCommerceApiConnectionException;
use WcMipConnector\Manager\LanguageReportManager;
use WcMipConnector\Manager\TaxesManager;

class ShippingService
{
    private const COST_CARRIER_VALUE = 10000;
    private const TAX_DIVISION = 100;
    private const SHIPPING_TAX_COST = 0;
    private const ROUND_PRECISION = 2;
    private const API_CARRIER_STATUS = 0;

    /** @var ApiController */
    protected $apiController;
    /** @var ApiController */
    protected $taxesManager;
    /** @var LanguageReportManager */
    protected $languageManager;

    public function __construct()
    {
        $this->apiController = new ApiController();
        $this->taxesManager = new TaxesManager();
        $this->languageManager = new LanguageReportManager();
    }

    /**
     * @param string $isoCountry
     * @param string $postCode
     * @param array $productArray
     *
     * @return array
     * @throws WooCommerceApiConnectionException
     */
    public function getPackageShippingCost(string $isoCountry, string $postCode, array $productArray): array
    {
        $carrierMostCheap = [];

        $requestParams = array_merge(
            [
                'iso_country' => $isoCountry,
                'postcode' => $postCode,
            ],

            $productArray
        );

        $language = explode('_', $this->languageManager->getDefaultLanguageIsoCode());
        $languageIsoCode = $language[1];
        $carriers = $this->apiController->send('costs', $requestParams, $languageIsoCode);

        if (array_key_exists('status', $carriers) && $carriers['status'] === self::API_CARRIER_STATUS) {
            $carrierMostCheap['msg'] = $carriers['msg'];

            return $carrierMostCheap;
        }

        if (!array_key_exists('shipping_costs', $carriers)) {
            return $carrierMostCheap;
        }

        $shippingTaxCost = self::COST_CARRIER_VALUE;

        foreach ($carriers['shipping_costs'] as $shippingCost) {
            if ($shippingCost['carrier_cost'] < (float)$shippingTaxCost) {
                $carrierMostCheap = $shippingCost;
                $taxes = $this->getShippingTaxByIsoCountry($isoCountry);
                if($taxes){
                    $taxDivided = $taxes[$isoCountry]['Rate'] / self::TAX_DIVISION;
                    $taxShippingRate = $shippingCost['carrier_cost'] * $taxDivided;
                    $shippingTaxCost = round($shippingCost['carrier_cost'] + $taxShippingRate, self::ROUND_PRECISION);
                    $carrierMostCheap['carrier_cost'] = $shippingTaxCost;
                    $carrierMostCheap['taxes'] = $shippingTaxCost - $shippingCost['carrier_cost'];
                    $carrierMostCheap['id'] = $taxes[$isoCountry]['Id'];
                    continue;
                }

                $carrierMostCheap['carrier_cost'] = round($shippingCost['carrier_cost'], self::ROUND_PRECISION);
            }
        }

        return $carrierMostCheap;
    }

    /**
     * @param string $isoCountry
     * @return array
     * @throws WooCommerceApiConnectionException
     */
    public function getShippingTaxByIsoCountry(string $isoCountry): array
    {
        $taxRateIndexedById = [];
        $taxes = $this->taxesManager->getTaxes();
        $shippingTaxCost = (float)self::SHIPPING_TAX_COST;

        if ($taxes){
            foreach ($taxes as $tax){
                if ((($tax['country'] === $isoCountry) || (empty($tax['country']) && empty($tax['state']) && empty($tax['postcode']) && empty($tax['city']))) && $shippingTaxCost < (float)$tax['rate']) {
                    $taxRateIndexedById[$isoCountry] = [
                        'Rate' => $tax['rate'],
                        'Id' => $tax['id']
                    ];
                    $shippingTaxCost = $tax['rate'];
                }
            }
        }

        return $taxRateIndexedById;
    }
}