<?php

namespace WcMipConnector\Service;

defined('ABSPATH') || exit;

use WcMipConnector\Enum\MipWcConnector;
use WcMipConnector\Manager\SystemManager;
use WcMipConnector\Manager\WoocommerceApiConnectionManager;
use WcMipConnector\Repository\SystemRepository;

class SystemService
{
    private const API_BASE_URI = 'https://platform.bigbuy.eu';
    private const API_CONNECTOR_TOKEN = 'MmQzMTk3YTg3NDRjZTNiNWY4YTMwZTMwNThiN2UyYTZiMzgyZDcwMTFjNTY3ZjkwNjg5OWQ3NzdiYWVlNzNjYQ';
    private const MODULE_DIR = __DIR__.'/../../';
    private const UPGRADE_SQL_DIR = '/upgrade/sql/';
    private const UPGRADE_DIR = '/upgrade/';
    private const SUCCESS = 200;
    private const PATRON = 'sql-upgrade-';
    private const FACTOR_GB_TO_MB = 1024.0;
    private const FACTOR_KB_TO_MB = 0.001;
    private const FACTOR_B_TO_MB = 0.000001;
    private const MIN_MEMORY_LIMIT = 1024;
    private const MIN_TIME_LIMIT = 15000;
    private const KB_VALUE = 1024;
    private const KB_MEDIUM_VALUE = 512;
    private const KB_LOW_VALUE = 256;
    private const NON_VALUE = 0;
    private const MEMORY_INFO_PATH = '/proc/meminfo';
    private const TOTAL_MEMORY = 'MemTotal';
    private const FREE_MEMORY = 'MemFree';
    private const BATCH_BY_DEFAULT = 100;
    private const MEDIUM_BATCH_VALUE = 80;
    private const LOW_BATCH_VALUE = 50;
    private const MINIMUM_BATCH_VALUE = 5;
    private const FLOCK_DIR = '/usr/bin/flock';

    /** @var SystemManager  */
    protected $manager;
    /** @var SystemRepository */
    protected $repository;
    /** @var LoggerService */
    protected $logger;

    /**
     * SystemService constructor.
     */
    public function __construct()
    {
        $this->manager = new SystemManager();
        $this->repository = new SystemRepository();
        $this->logger = new LoggerService();
    }

    public function checkIfPluginNeedUpgrade(string $fileVersion): bool
    {
        if(!$this->isNeedUpdateModule($fileVersion)) {
            return true;
        }

        if (!$this->upgradeModuleFiles()) {
            return false;
        }

        $mipTables = $this->manager->getMipTablesFromDatabase();
        $this->repository->createWcMipConnectorTables($mipTables);

        if(!$this->loadUpgradeSqlFromFiles($fileVersion)) {
            return false;
        }

        return true;
    }

    /**
     * @param string $fileVersion
     *
     * @return bool
     */
    private function isNeedUpdateModule(string $fileVersion): bool
    {
        $moduleVersion = $this->manager->getModuleVersion();

        return 0 < version_compare($fileVersion, $moduleVersion);
    }

    private function upgradeModuleFiles(): bool
    {
        try {
            $nameZipTmp = 'tmpUpdateModule.zip';
            $url = self::API_BASE_URI.'/rest/selling_channel_connector/2.json';
            $headers = array(
                'Accept' => 'application/json',
                'Content-Type' => 'application/json',
                'Authorization' => 'Bearer '.self::API_CONNECTOR_TOKEN
            );
            $args = array(
                'timeout' => '60',
                'method' => 'GET',
                'httpversion' => '1.1',
                'headers' => $headers
            );
            $response = wp_remote_post($url, $args);
            $statusCode = wp_remote_retrieve_response_code($response);

            if ($statusCode !== self:: SUCCESS) {
                return false;
            }

            $connectorInfo = json_decode(wp_remote_retrieve_body($response));
            $connectorUrl = $connectorInfo->url;
            $connectorVersion = $connectorInfo->version;
            $moduleVersion = $this->manager->getModuleVersion();

            if (version_compare($connectorVersion, $moduleVersion) <= 0) {
                return false;
            }

            $updateContent = file_get_contents($connectorUrl);

            if (!$updateContent) {
                return false;
            }

            file_put_contents(self::MODULE_DIR.'/'.$nameZipTmp, $updateContent);
            $zip = new \ZipArchive();
            $result = $zip->open(self::MODULE_DIR.'/'.$nameZipTmp);

            if (!$result) {
                return false;
            }

            $zip->extractTo(self::MODULE_DIR.'/../');
            $zip->close();
            unlink(self::MODULE_DIR.'/'.$nameZipTmp);
            $this->updateModuleVersion($connectorVersion);
            return true;
        } catch (\Exception $e) {
            return $e->getMessage().$e->getTrace();
        }
    }

    public function loadUpgradeSqlFromFiles($versionFile): ?bool
    {
        if (file_exists(self::MODULE_DIR.self::UPGRADE_SQL_DIR) && ($files = scandir(self::MODULE_DIR.self::UPGRADE_SQL_DIR))) {
            foreach ($files as $file) {
                $sql = [];
                if (!in_array($file, array('.', '..', '.svn', 'index.php')) && preg_match('/\.php$/', $file)) {
                    $versionUpgrade = $this->getVersionFromFileNames($file);
                    if ($versionUpgrade >= $versionFile) {
                        $sql = require self::MODULE_DIR.self::UPGRADE_SQL_DIR.$file;
                        if (empty($sql) || !is_array($sql)) {
                            return false;
                        }
                        foreach ($sql as $query) {
                            $result = $this->manager->executeSql($query);
                            $require = require self::MODULE_DIR.self::UPGRADE_DIR.'upgrade-'.$versionUpgrade.'.php';
                            if($result && $require) {
                                $this->updateModuleVersion($versionUpgrade);
                            }
                        }
                    }
                }
            }
        }
        return true;
    }

    private function getVersionFromFileNames($file): string
    {
        $version = str_replace('-','.',str_replace('.php','',str_replace(self::PATRON,'', $file)));

        return $version;
    }

    /**
     * @param string $fileVersion
     * @return bool
     */
    private function updateModuleVersion(string $fileVersion): bool
    {
        return $this->manager->updateModuleVersion($fileVersion);
    }

    /**
     * @param $code
     * @param string $message
     * @param array $data
     */
    public function JsonResponse($code, $message = '', array $data = []): void
    {
        $result['Code'] = $code;
        $result['Message'] = $message;
        $result['Data'] = $data;

        header('Content-Type: application/json');

        $response = json_encode($result, JSON_UNESCAPED_UNICODE);

        if ($code > 399) {
            //MMOLogger::getInstance()->alert("$code - $message RESPONSE: $response");
        }
        else {
            //MMOLogger::getInstance()->debug("$code - $message RESPONSE: $response");
        }

        die($response);
    }

    public function resetExecutionLimits(): void
    {
        $maxExecutionTime = @ini_get('max_execution_time');
        if ((int)$maxExecutionTime > 0 && (int)$maxExecutionTime < self::MIN_TIME_LIMIT) {
            set_time_limit(self::MIN_TIME_LIMIT);
        }

        $memoryLimit = $this->convertShorthandBytesToMegabytes(@ini_get('memory_limit'));
        if ($memoryLimit > 0 && $memoryLimit < self::MIN_MEMORY_LIMIT) {
            ini_set('memory_limit', self::MIN_MEMORY_LIMIT.'M');
        }
    }

    public function enablePlugin(): bool
    {
        $isEnable = is_plugin_active(MipWcConnector::MODULE_NAME.'/WcMipconnector.php');
        if ($isEnable) {
            return true;
        }

        $hasBeenActivated = activate_plugin(MipWcConnector::MODULE_NAME.'/WcMipconnector.php');
        if (is_wp_error($hasBeenActivated)) {
            return false;
        }

        return true;
    }

    private function convertShorthandBytesToMegabytes(string $shorthandValue): int
    {
        $sanitizedShorthandValue = mb_strtolower(trim($shorthandValue));
        $value = (int)$sanitizedShorthandValue;

        if (false !== mb_strpos($sanitizedShorthandValue, 'm')) {
            return $value;
        }

        if (false !== mb_strpos($sanitizedShorthandValue, 'g')) {
            return (int)($value * self::FACTOR_GB_TO_MB);
        }

        if (false !== strpos($sanitizedShorthandValue, 'k')) {
            return (int)($value * self::FACTOR_KB_TO_MB);
        }

        return (int)($value * self::FACTOR_B_TO_MB);
    }

    /**
     * @param \WC_Emails $emails
     * @return void
     */
    public function updateSendOrderEmailConfiguration(\WC_Emails $emails): void
    {
        $woocommerceApiConnectionManager = new WoocommerceApiConnectionManager();
        $sendEmailStatus = $woocommerceApiConnectionManager->getSendEmailConfiguration();

        if ($sendEmailStatus) {
            $this->enableOrderEmailHooks($emails);
            return;
        }

        $this->disableOrderEmailHooks($emails);
    }

    /**
     * @param \WC_Email_New_Order $emails
     */
    public function disableOrderEmailHooks(\WC_Emails $emails): void
    {
        remove_action('woocommerce_order_status_completed_notification', array($emails->emails['WC_Email_Customer_Completed_Order'], 'trigger'));
        remove_action('woocommerce_order_status_pending_to_processing_notification', array($emails->emails['WC_Email_Customer_Processing_Order'], 'trigger'));
        remove_action('woocommerce_order_status_pending_to_on-hold_notification', array($emails->emails['WC_Email_Customer_Processing_Order'], 'trigger'));
    }

    /**
     * @param \WC_Emails $emails
     */
    public function enableOrderEmailHooks(\WC_Emails $emails): void
    {
        add_action('woocommerce_order_status_completed_notification', array($emails, 'trigger'));
        add_action('woocommerce_order_status_pending_to_processing_notification', array($emails, 'trigger'));
        add_action('woocommerce_order_status_pending_to_on-hold_notification', array($emails, 'trigger'));
    }

    /**
     * @return array
     */
    public function getMemoryInfo(): array
    {
        $memoryInfo = [];

        if (stripos(PHP_OS_FAMILY, 'Windows') !== false) {
            $cmd = 'wmic ComputerSystem get TotalPhysicalMemory';
            @exec($cmd, $outputTotalPhysicalMemory);
            $cmd = 'wmic OS get FreePhysicalMemory';
            @exec($cmd, $outputFreePhysicalMemory);

            $memoryInfo[self::TOTAL_MEMORY] = self::bytesToMegabytes($outputTotalPhysicalMemory[1]);
            $memoryInfo[self::FREE_MEMORY] = self::bytesToMegabytes($outputFreePhysicalMemory[1]);
        }

        if (is_readable('/proc/meminfo')) {
            $memoryInfoData = explode("\n", file_get_contents(self::MEMORY_INFO_PATH));
            foreach ($memoryInfoData as $line) {
                $key = strtok($line, ':');
                if($key === self::TOTAL_MEMORY || $key === self::FREE_MEMORY){
                    $memoryInfo[$key] = (int)(filter_var($line, FILTER_SANITIZE_NUMBER_INT) / self::KB_VALUE);
                }
            }
        }

        return $memoryInfo;
    }

    /**
     * @return int
     */
    public function setBatchValue(): int
    {
        $memoryInfo = $this->getMemoryInfo();

        if ($memoryInfo[self::FREE_MEMORY] === self::NON_VALUE){
            return self::MINIMUM_BATCH_VALUE;
        }

        if ($memoryInfo[self::FREE_MEMORY] <= self::KB_LOW_VALUE){
            return self::LOW_BATCH_VALUE;
        }

        if ($memoryInfo[self::FREE_MEMORY] <= self::KB_MEDIUM_VALUE){
            return self::MEDIUM_BATCH_VALUE;
        }

        return self::BATCH_BY_DEFAULT;
    }

    /**
     * @return int
     */
    public function getBatchValue(): int
    {
        $batchValue = $this->setBatchValue();
        $this->logger->info('Batch value: '.$batchValue);

        return $batchValue;
    }

    /**
     * @return array
     */
    public function getStorageServerInfo(): array
    {
        $dataStorage = [];
        $dataStorage['Total'] = self::bytesToGigabytes(disk_total_space(MipWcConnector::MODULE_DIR));
        $dataStorage['Available'] = self::bytesToGigabytes(disk_free_space(MipWcConnector::MODULE_DIR));

        return $dataStorage;
    }

    public static function bytesToGigabytes($bytes)
    {
        return round($bytes / 1024 / 1024 / 1024, 2);
    }

    public static function bytesToMegabytes($bytes)
    {
        return round($bytes / 1024 / 1024, 2);
    }

    /**
     * @return bool
     */
    public function checkFLockExists(): bool
    {
        if(file_exists(self::FLOCK_DIR)){
            return true;
        }

        return false;
    }
}