<?php

namespace WcMipConnector\Service;

defined('ABSPATH') || exit;

use WcMipConnector\Exception\WooCommerceApiConnectionException;
use WcMipConnector\Factory\TagMapFactory;
use WcMipConnector\Factory\TagFactory;
use WcMipConnector\Manager\TagManager;
use WcMipConnector\Manager\TagMapManager;

class TagService
{
    /** @var TagFactory  */
    protected $tagFactory;
    /** @var TagMapFactory  */
    protected $tagMapFactory;
    /** @var TagMapManager  */
    protected $tagMapManager;
    /** @var TagManager  */
    protected $tagManager;
    /** @var ImportProcessTagService  */
    protected $importProcessService;
    /** @var LoggerService  */
    protected $loggerService;
    /** @var LoggerService  */
    protected $logger;
    /** @var array */
    protected $tagsToDelete;
    /** @var array */
    protected $languages;
    /** @var SystemService */
    protected $systemService;

    public function __construct()
    {
        $this->tagFactory = new TagFactory();
        $this->tagMapFactory = new TagMapFactory();
        $this->tagMapManager = new TagMapManager();
        $this->tagManager = new TagManager();
        $this->importProcessService = new ImportProcessTagService();
        $this->loggerService = new LoggerService();
        $this->logger = $this->loggerService->getInstance();
        $this->systemService = new SystemService();
    }

    /**
     * @param array $tags
     * @param int $fileId
     * @param array $languages
     * @return array
     */
    public function processByBatch(array $tags, int $fileId, array $languages): array
    {
        $this->languages = $languages;
        $tagsIndexedByTagMapId = [];
        $tagIdsToUpdateIndexedByTagMapId = [];
        $tagsShopIndexedByTagId = [];
        $tagShopIndexedByMapIds = [];
        $batches = array_chunk($tags, $this->systemService->getBatchValue(), true);
        $tagsProcessedArray = [[]];

        foreach ($batches as $tagsToProcess) {
            $this->tagsToDelete = [];
            $tagFactorySave = [];
            $tagFactoryToUpdate = [];
            $tagsToSaveIndexedByTagMapId = [];
            $tagsToUpdateIndexedByTagMapId = [];
            $tagsToNotProcess = [];

            foreach ($tagsToProcess as $tagData) {
                $tagFactory = $this->tagFactory->create($tagData, $languages);
                $tagsIndexedByTagMapId[$tagData['TagID']] = $tagFactory;
            }

            $tagMapsIndexedByTagMapId = $this->tagMapManager->findByTagMapIdsIndexedByTagMapId(array_keys($tagsToProcess));
            $tagVersionsIndexedByTagMapId = $this->tagMapManager->findVersionsIndexedByTagMapId(array_keys($tagMapsIndexedByTagMapId));

            if (!empty($tagMapsIndexedByTagMapId)) {
                $tagsShopIndexedByTagId = $this->tagManager->findByTagShopIdsIndexedByTagId(array_values($tagMapsIndexedByTagMapId));
            }

            if($tagsShopIndexedByTagId){
                foreach ($tagMapsIndexedByTagMapId as $tagParentMappedId => $tagMapId){
                    if(array_key_exists($tagMapId, $tagsShopIndexedByTagId)){
                        $tagShopIndexedByMapIds[$tagParentMappedId] = $tagMapId;
                    }
                }
            }

            foreach ($tagsIndexedByTagMapId as $tagMapId => $tagMapData) {
                if (array_key_exists($tagMapId, $tagShopIndexedByMapIds)){
                    $version = json_decode($tagVersionsIndexedByTagMapId[$tagMapId], true);
                    $tagsIndexedByIsoCode = array_column($tagsToProcess[$tagMapId]['TagLangs'], 'Version', 'IsoCode');
                    foreach ($version as $isoCode => $codeVersion){
                        if (array_key_exists($tagMapId, $tagsIndexedByTagMapId)
                            && $codeVersion > $tagsIndexedByIsoCode[$isoCode]){
                            $tagsToNotProcess[$tagMapId] = true;
                            $tagsProcessedArray[][] =  ['shopId' => (int)$tagShopIndexedByMapIds[$tagMapId], 'id' => (int)$tagMapId];
                            continue;
                        }
                    }

                    if (array_key_exists($tagMapId, $tagsToNotProcess)){
                        continue;
                    }

                    $tagMapData->id = $tagShopIndexedByMapIds[$tagMapId];
                    $tagIdsToUpdateIndexedByTagMapId[$tagMapId] = false;
                    $tagFactoryToUpdate[] = $tagMapData;
                    $tagsToUpdateIndexedByTagMapId[$tagMapId] = $tagMapData;
                    continue;
                }

                if (!array_key_exists($tagMapId, $tagMapsIndexedByTagMapId)) {
                    $tagIdsToUpdateIndexedByTagMapId[$tagMapId] = false;
                    $tagFactorySave[] = $tagMapData;
                    $tagsToSaveIndexedByTagMapId[$tagMapId] = $tagMapData;
                    continue;
                }

                $tagsToSaveIndexedByTagMapId[$tagMapId] = $tagMapData;
                $tagIdsToUpdateIndexedByTagMapId[$tagMapId] = true;
                $tagFactorySave[] = $tagMapData;
            }

            unset($tagsIndexedByTagMapId);

            $tagRequest = ['create' => $tagFactorySave, 'update' => $tagFactoryToUpdate];

            try {
                $tagsBatchResponse = $this->tagManager->save($tagRequest);
            } catch (WooCommerceApiConnectionException $e) {
                $this->logger->error('Tag Save - Batch Request: '.json_encode($tagRequest, true).' - Exception Message: '.$e->getMessage());
                continue;
            }

            if (array_key_exists('create', $tagsBatchResponse) && $tagsBatchResponse['create']) {
                $tagsProcessedArray[] = $this->save(
                    $tagsBatchResponse['create'],
                    $tagsToSaveIndexedByTagMapId,
                    $tagIdsToUpdateIndexedByTagMapId,
                    $tagsToProcess,
                    $fileId
                );
            }

            if (array_key_exists('update', $tagsBatchResponse) && $tagsBatchResponse['update']) {
                $tagsProcessedArray[] = $this->update(
                    $tagsBatchResponse['update'],
                    $tagsToUpdateIndexedByTagMapId,
                    $tagsToProcess,
                    $fileId
                );
            }

            if (!empty($this->tagsToDelete)) {
                $this->delete($this->tagsToDelete,$fileId);
            }
        }

        $tagsProcessed = array_merge(...$tagsProcessedArray);

        return array_column($tagsProcessed, 'shopId', 'id');
    }

    /**
     * @param array $tagsBatchResponse
     * @param array  $tagsIndexedByTagMapId
     * @param array  $tagIdsToUpdateIndexedByTagMapId
     * @param array  $tags
     * @param int    $fileId
     *
     * @return array
     */
    private function save(array $tagsBatchResponse, array $tagsIndexedByTagMapId, array $tagIdsToUpdateIndexedByTagMapId, array $tags, int $fileId): array
    {
        $tagIds = [];
        $tagMapIndex = 0;

        foreach ($tagsIndexedByTagMapId as $tagId => $tagMapData) {
            $tag = $tagsBatchResponse[$tagMapIndex];
            $tagMapIndex++;

            if (!$tag) {
                continue;
            }

            if (array_key_exists('error', $tag) && $tag['error']) {
                $importProcess = $this->importProcessService->setFailure($tagId, $fileId);
                $this->logger->error('Save Tag Id: '.$tagId.' in the File Id: '.$fileId.' - Batch Error Message: '.$tag['error']['message'].' - Tag object: '.json_encode($tagMapData, true). ' - SetFailure result: '.$importProcess);
                continue;
            }

            $tagToMap = $this->tagMapFactory->create($tags[$tagId], $tag['id'], $this->languages);

            if (array_key_exists($tagId, $tagIdsToUpdateIndexedByTagMapId) && $tagIdsToUpdateIndexedByTagMapId[$tagId]) {
                $tagMapped = $this->tagMapManager->update($tag['id'], $tagToMap);
            } else {
                $tagMapped = $this->tagMapManager->save($tagToMap);
            }

            $importProcess = $this->importProcessService->setSuccess($tagId, $fileId);

            if (!$tagMapped) {
                $this->logger->error('Map in Save Tag Id: '.$tagId.' in the File Id: '.$fileId.' - Tag Shop Id: '.$tag['id'].' - Data to map: '.json_encode($tagToMap, true). ' - SetSuccess result: '.$importProcess);
                $this->tagsToDelete[] = $tag['id'];
                continue;
            }

           if (!$importProcess) {
               $this->logger->error('SetSuccess in Save Tag Id: '.$tagId.' in the File Id: '.$fileId.' - SetSuccess result: '.$importProcess);
           }

            $tagIds[] = ['shopId' => $tag['id'], 'id' => $tagId];
        }

        return $tagIds;
    }

    /**
     * @param array $tagsBatchResponse
     * @param array $tagsIndexedByTagMapId
     * @param array $tags
     * @param int   $fileId
     *
     * @return array
     */
    private function update(array $tagsBatchResponse, array $tagsIndexedByTagMapId, array $tags, int $fileId): array
    {
        $tagIds = [];
        $tagMapIndex = 0;

        foreach ($tagsIndexedByTagMapId as $tagId => $tagMapData) {
            $tag = $tagsBatchResponse[$tagMapIndex];
            $tagMapIndex++;

            if (!$tag) {
                continue;
            }

            if (array_key_exists('error', $tag) && $tag['error']) {
                $importProcess = $this->importProcessService->setFailure($tagId, $fileId);
                $this->logger->error('Update Tag Id: '.$tagId.' in the File Id: '.$fileId.' - Batch Error Message: '.$tag['error']['message'].' - Tag object: '.json_encode($tagMapData, true). ' - SetFailure result: '.$importProcess);
                continue;
            }

            $tagToMap = $this->tagMapFactory->create($tags[$tagId], $tag['id'], $this->languages);
            $tagMapped = $this->tagMapManager->update($tag['id'], $tagToMap);
            $importProcess = $this->importProcessService->setSuccess($tagId, $fileId);

            if (!$tagMapped) {
                $this->logger->error('Map in Update Tag Id: '.$tagId.' in the File Id: '.$fileId.' - Tag Shop Id: '.$tag['id'].' - Data to map: '.json_encode($tagToMap, true). ' - SetSuccess result: '.$importProcess);
            }

            if (!$importProcess) {
                $this->logger->error('SetSuccess in Update Tag Id: '.$tagId.' in the File Id: '.$fileId.' - SetSuccess result: '.$importProcess);
            }

            $tagIds[] = ['shopId' => $tag['id'], 'id' => $tagId];
        }

        return $tagIds;
    }

    /**
     * @param array $tagsToDelete
     * @param int $fileId
     */
    private function delete(array $tagsToDelete, int $fileId): void
    {
        $tagData = ['delete' => $tagsToDelete];

        try {
            $deleteTagsBatchResponse = $this->tagManager->delete($tagData);
        } catch (WooCommerceApiConnectionException $e) {
            $this->logger->error('Tag Delete of File Id: '.$fileId.' - Batch Request: '.json_encode($tagData, true).' - Exception Message: '.$e->getMessage());
            return;
        }

        foreach ($deleteTagsBatchResponse['delete'] as $tag) {
            if (array_key_exists('error', $tag) && $tag['error']) {
                $this->logger->error('Delete of shop unmapped tag: '. $tag['id'] . ' - Error Message: '.$tag['error']['message']);
            }
        }
    }
}