<?php

namespace WcMipConnector\Service;

use WcMipConnector\Factory\UnPublishFactory;
use WcMipConnector\Manager\ProductManager;
use WcMipConnector\Manager\ProductMapManager;
use WcMipConnector\Factory\ProductMapFactory;

defined('ABSPATH') || exit;

class UnPublishProductService
{
    /** @var ProductManager */
    protected $productManager;
    /** @var ProductMapManager */
    protected $productMapManager;
    /** @var UnPublishFactory */
    protected $disableFactory;
    /** @var ImportProcessProductService */
    protected $importProcessService;
    /** @var LoggerService */
    protected $logger;
    /** @var ProductMapFactory */
    protected $productMapFactory;
    /** @var SystemService */
    protected $systemService;

    public function __construct()
    {
        $this->productManager = new ProductManager();
        $this->productMapManager = new ProductMapManager();
        $this->disableFactory = new UnPublishFactory;
        $this->importProcessService = new ImportProcessProductService();
        $this->logger = new LoggerService;
        $this->productMapFactory = new ProductMapFactory;
        $this->systemService = new SystemService();
    }

    /**
     * @param array $productsDisabled
     * @param int   $fileId
     *
     * @throws \Exception
     */
    public function disableByBatch(array $productsDisabled, int $fileId):void
    {
        $productShopIds = [];
        $productShopIdsIndexedByProductMapId = [];

        if (!$productsDisabled){
            $productsDisabled = [];
        }

        $batches = array_chunk($productsDisabled, $this->systemService->getBatchValue(), true);
        foreach ($batches as $batchToProcess) {
            $productFactoryToUpdate = [];
            $productToUpdate = [];

            $productMapIdsIndexedByProductMapId = $this->productMapManager->findByProductMapIdsIndexedByProductMapId(array_keys($productsDisabled));

            if(!empty($productMapIdsIndexedByProductMapId)){
                $productShopIds = $this->productManager->findByProductShopIdsIndexedByProductId(array_values($productMapIdsIndexedByProductMapId));
            }

            if($productShopIds){
                foreach ($productMapIdsIndexedByProductMapId as $productParentIdMapped => $productIdMapped){
                    if(array_key_exists($productIdMapped, $productShopIds)){
                        $productShopIdsIndexedByProductMapId[$productParentIdMapped] = $productIdMapped;
                    }
                }
            }

            foreach ($batchToProcess as $productMapId => $productMapData) {
                if (array_key_exists($productMapId, $productShopIdsIndexedByProductMapId)) {
                    $disableData = $this->disableFactory->create($productMapData, $productShopIdsIndexedByProductMapId[$productMapId]);
                    $productFactoryToUpdate[] = $disableData;
                    $productToUpdate[$productMapId] = $disableData;
                    continue;
                }

                $this->logger->getInstance()->error('Product '.$productMapId.' of fileID '.$fileId.' has been not disabled because it doesnt exist in the shop.');
            }

            $productRequest = ['update' => $productFactoryToUpdate];
            $productBatchResponse = $this->productManager->save($productRequest);

            if (array_key_exists('update', $productBatchResponse) && $productBatchResponse['update']) {
                $this->update($productBatchResponse['update'], $productToUpdate, $fileId);
            }
        }
    }

    /**
     * @param array $productBatchResponse
     * @param array $productToUpdate
     * @param int   $fileId
     *
     * @return void
     * @throws \Exception
     */
    private function update(array $productBatchResponse,  array $productToUpdate, int $fileId): void
    {
        $productMapIndex = 0;
        foreach ($productToUpdate as $productId => $productMapData){
            $product = $productBatchResponse[$productMapIndex];
            $productMapIndex++;
            if(!$product){
                continue;
            }

            if(array_key_exists('error', $product) && $product['error']){
                $importProcess = $this->importProcessService->setFailure($productId, $fileId);
                $this->logger->getInstance()->error('Product '.$productId.' of fileID '.$fileId.' has been not disabled. Batch response: '
                    .$product['error']['message'].'. Product array: '.json_encode($productMapData).'. ImportProcess added: '.$importProcess);
                continue;
            }

            $productToMap = $this->productMapFactory->createUnPublish($productId, $product['id']);
            $productUpdated = $this->productMapManager->update($product['id'], $productToMap);
            $importProcess = $this->importProcessService->setSuccess($productId, $fileId);
            if(!$productUpdated){
                $this->logger->getInstance()->error('Product '.$productId.' of fileID '.$fileId.' has been not disabled. Product array: '
                    .json_encode($productToMap).' with shopID: '.$product['id'].'. ImportProcess added: '.$importProcess);
            }
        }
    }
}