<?php

namespace WcMipConnector\Service;

defined('ABSPATH') || exit;

use WcMipConnector\Factory\VariationFactory;
use WcMipConnector\Factory\VariationMapFactory;
use WcMipConnector\Manager\VariationManager;
use WcMipConnector\Manager\VariationMapManager;

class VariationService
{
    /** @var VariationManager */
    protected $variationManager;
    /** @var VariationMapManager */
    protected $variationMapManager;
    /** @var VariationFactory */
    protected $variationFactory;
    /** @var VariationMapFactory */
    protected $variationMapFactory;
    /** @var ImportProcessVariationService */
    protected $importProcessService;
    /** @var LoggerService */
    protected $logger;
    /** @var SystemService */
    protected $systemService;

    public function __construct()
    {
        $this->variationManager = new VariationManager();
        $this->variationMapManager = new VariationMapManager();
        $this->variationFactory = new VariationFactory();
        $this->variationMapFactory = new VariationMapFactory();
        $this->importProcessService = new ImportProcessVariationService();
        $this->logger = new LoggerService();
        $this->systemService = new SystemService();
    }

    /**
     * @param array $variationsIndexedByVariationMapId
     * @param int   $variationParentId
     * @param array $variation
     * @param int   $fileId
     *
     * @return array
     */
    public function processVariationByBatch(array $variationsIndexedByVariationMapId = null, int $variationParentId = null, array $variation = null, int $fileId = null): array
    {
        $processedArray = [[]];

        if (!$variationParentId){
            foreach ($variationsIndexedByVariationMapId as $variationId => $variationData){
                $this->importProcessService->setFailure($variationId, $fileId);
                $this->logger->getInstance()->error('VariationID: '.$variationId.' of FileID: '.$fileId.' does not have parent');
            }

            return [];
        }

        $batches = array_chunk($variationsIndexedByVariationMapId, $this->systemService->getBatchValue(), true);
        foreach ($batches as $batchToProcess) {
            $variationShopIndexedByVariationId = [];
            $variationFactorySave = [];
            $variationFactoryToUpdate = [];
            $variationIdsToUpdateIndexedByVariationMapId = [];
            $variationToSaveIndexedByVariationMapId = [];
            $variationToUpdateIndexedByVariationMapId = [];

            $variationMapsIndexedByVariationMapId = $this->variationMapManager->findByVariationMapIdsIndexedByVariationMapId(array_keys($batchToProcess));

            if (!empty($variationMapsIndexedByVariationMapId)) {
                $variationShopIndexedByVariationId = $this->variationManager->findByVariationMapIdsIndexedByVariationId(array_values($variationMapsIndexedByVariationMapId));
            }

            foreach ($variationsIndexedByVariationMapId as $variationMapId => $variationMapData) {
                if (!array_key_exists($variationMapId, $variationMapsIndexedByVariationMapId)) {
                    $variationIdsToUpdateIndexedByVariationMapId[$variationMapId] = false;
                    $variationFactorySave[] = $variationMapData;
                    $variationToSaveIndexedByVariationMapId[$variationMapId] = $variationMapData;
                    continue;
                }

                if (array_key_exists($variationMapId, $variationMapsIndexedByVariationMapId) && array_key_exists(
                        $variationMapsIndexedByVariationMapId[$variationMapId],
                        $variationShopIndexedByVariationId
                    )) {
                    $variationMapData->id = $variationMapsIndexedByVariationMapId[$variationMapId];
                    $variationIdsToUpdateIndexedByVariationMapId[$variationMapId] = false;
                    $variationFactoryToUpdate[] = $variationMapData;
                    $variationToUpdateIndexedByVariationMapId[$variationMapId] = $variationMapData;
                    continue;
                }

                $variationToSaveIndexedByVariationMapId[$variationMapId] = $variationMapData;
                $variationIdsToUpdateIndexedByVariationMapId[$variationMapId] = true;
                $variationFactorySave[] = $variationMapData;
            }

            $variationRequest = ['create' => $variationFactorySave, 'update' => $variationFactoryToUpdate];
            $variationBatchResponse = $this->variationManager->save($variationRequest, $variationParentId);
            if (array_key_exists('create', $variationBatchResponse) && $variationBatchResponse['create']) {
                $processedArray[] = $this->save(
                    $variationBatchResponse['create'],
                    $variationToSaveIndexedByVariationMapId,
                    $variationIdsToUpdateIndexedByVariationMapId,
                    $variation,
                    $fileId
                );
            }

            if (array_key_exists('update', $variationBatchResponse) && $variationBatchResponse['update']) {
                $processedArray[] = $this->update(
                    $variationBatchResponse['update'],
                    $variationToUpdateIndexedByVariationMapId,
                    $variation,
                    $fileId
                );
            }
        }

        $processed = array_merge(...$processedArray);

        return array_column($processed, 'shopId', 'id');
    }

    /**
     * @param array $variationBatchResponse
     * @param array  $variationIndexedByVariationMapId
     * @param array  $variationIdsToUpdateIndexedByVariationMapId
     * @param array  $variations
     * @param int    $fileId
     *
     * @return array
     */
    private function save(array $variationBatchResponse, array $variationIndexedByVariationMapId, array $variationIdsToUpdateIndexedByVariationMapId, array $variations, int $fileId): array
    {
        $variationIds = [];
        $variationMapIndex = 0;

        foreach ($variationIndexedByVariationMapId as $variationId => $variationMapData){
            $variation = $variationBatchResponse[$variationMapIndex];
            $variationMapIndex++;
            $variationUpdated = true;
            $variationSaved = true;

            if(!$variation){
                continue;
            }

            if(array_key_exists('error', $variation) && $variation['error']){
                $importProcess = $this->importProcessService->setFailure($variationId, $fileId);
                $this->logger->getInstance()->error('VariationID: '.$variationId.' of FileID: '.$fileId.' has been not created. Batch response: '
                    .$variation['error']['message'].' - Variation SKU: '.$variationMapData->sku.' - ImportProcess mapped: '.$importProcess);
                continue;
            }

            $variationToMap = $this->variationMapFactory->create($variations[$variationId], $variation['id']);

            if(array_key_exists($variationId, $variationIdsToUpdateIndexedByVariationMapId) && $variationIdsToUpdateIndexedByVariationMapId[$variationId]){
                $variationUpdated = $this->variationMapManager->update($variation['id'], $variationToMap);
            }else{
                $variationSaved = $this->variationMapManager->save($variationToMap);
            }

            $importProcess = $this->importProcessService->setSuccess($variationId, $fileId);

            if (!$variationUpdated){
                $this->logger->getInstance()->error('VariationID: '.$variationId.' of FileID: '.$fileId.' has been not updated in the map - Variation SKU: '.$variationMapData->sku.' with ShopID: '.$variation['id'].' - ImportProcess mapped: '.$importProcess);
            }

            if (!$variationSaved){
                $this->logger->getInstance()->error('VariationID: '.$variationId.' of FileID: '.$fileId.' has been not created in the map - Variation SKU: '.$variationMapData->sku.' - ImportProcess mapped: '.$importProcess);
            }

            $variationIds[] = ['shopId' => $variation['id'], 'id' => $variationId];
        }

        return $variationIds;
    }

    /**
     * @param array $variationBatchResponse
     * @param array $variationIndexedByVariationMapId
     * @param array $variations
     * @param int   $fileId
     *
     * @return array
     */
    private function update(array $variationBatchResponse, array $variationIndexedByVariationMapId, array $variations, int $fileId): array
    {
        $variationIds = [];
        $variationMapIndex = 0;

        foreach ($variationIndexedByVariationMapId as $variationId => $variationMapData){
            $variation = $variationBatchResponse[$variationMapIndex];
            $variationMapIndex++;

            if(!$variation){
                continue;
            }

            if(array_key_exists('error', $variation) && $variation['error']){
                $importProcess = $this->importProcessService->setFailure($variationId, $fileId);
                $this->logger->getInstance()->error('VariationID: '.$variationId.' of FileID: '.$fileId.' has been not updated. Batch response: '
                    .$variation['error']['message'].' - Variation SKU: '.$variationMapData->sku.' - ImportProcess mapped: '.$importProcess);
                continue;
            }

            $variationToMap = $this->variationMapFactory->create($variations[$variationId], $variation['id']);
            $variationUpdated = $this->variationMapManager->update($variation['id'], $variationToMap);
            $importProcess = $this->importProcessService->setSuccess($variationId, $fileId);
            if (!$variationUpdated){
                $this->logger->getInstance()->error('VariationID: '.$variationId.' of FileID: '.$fileId.' has been not updated in the map - Variation SKU: '.$variationMapData->sku.' with ShopID: '.$variation['id'].' - ImportProcess mapped: '.$importProcess);
            }

            $variationIds[] = ['shopId' => $variation['id'], 'id' => $variationId];
        }

        return $variationIds;
    }
}