<?php

namespace WcMipConnector\Service;

require_once __DIR__.'/../../app/views/CarrierViewManager.php';

use WcMipConnector\Exception\NoResultException;
use WcMipConnector\Manager\WoocommerceApiConnectionManager;
use WcMipConnector\Manager\ZoneManager;
use WcMipConnector\Factory\ZoneFactory;
use WcMipConnector\Manager\ProductManager;
use WcMipConnector\Exception\WooCommerceApiConnectionException;
use WcMipConnector\Factory\MethodFactory;

class ZoneService
{
    private const VARIATION_ID = 0;
    private const METHOD_POSITION = 0;

    /**@var WoocommerceApiConnectionService */
    protected $wooCommerce;
    /** @var ZoneManager */
    protected $zoneManager;
    /** @var ZoneFactory */
    protected $zoneFactory;
    /** @var MethodFactory */
    protected $methodFactory;
    /** @var ZoneFactory */
    protected $carrierView;
    /** @var ZoneFactory */
    protected $productManager;
    /** @var ShippingService */
    protected $shippingService;
    /** @var LoggerService */
    protected $logger;

    public function __construct()
    {
        global $woocommerce;
        $this->wooCommerce = $woocommerce;
        $this->zoneManager = new ZoneManager();
        $this->zoneFactory = new ZoneFactory();
        $this->methodFactory = new MethodFactory();
        $this->carrierView = new \CarrierViewManager();
        $this->productManager = new ProductManager();
        $this->shippingService = new ShippingService();
        $loggerService = new LoggerService();
        $this->logger = $loggerService->getInstance();
    }

    /**
     * @return array
     * @throws WooCommerceApiConnectionException
     */
    public function findAllZones(): array
    {
        $findAllZones = $this->zoneManager->findAllZones();
        $zonesIndexedById = array_column($findAllZones, 'id', 'name');
        if(!array_key_exists(ZoneFactory::ALL_WORLD_ZONE, $zonesIndexedById)){
            $zoneModel = $this->zoneFactory->create();
            $zones = $this->zoneManager->createZones(json_decode(json_encode($zoneModel), true));

            return $this->findShippingZoneByZoneId($zones['id']);
        }

        return $this->findShippingZoneByZoneId($zonesIndexedById[ZoneFactory::ALL_WORLD_ZONE]);
    }

    /**
     * @param int $zoneId
     *
     * @return array
     * @throws WooCommerceApiConnectionException
     */
    public function findShippingZoneByZoneId(int $zoneId): array
    {
        $listCarriers = $this->zoneManager->findShippingZoneByZoneId($zoneId);
        $carriers = array_column($listCarriers, 'instance_id', 'method_id');
        if(!array_key_exists(ZoneFactory::LOCAL_PICKUP, $carriers)){
            $methodModel = $this->methodFactory->setLocalPickupMethod();
            $listCarriers = $this->zoneManager->createShippingZones($zoneId, json_decode(json_encode($methodModel), true));
        }

        return $listCarriers;
    }

    /**
     * @param array $shippingRates
     * @return array
     */
    public function shippingMethodsAvailable(array $shippingRates): array
    {
        $woocommerceApiConnectionManager = new WoocommerceApiConnectionManager();
        $bigBuyCarrierOption = $woocommerceApiConnectionManager->getBigBuyCarrierOption();
        $shippingMethods = array();

        foreach ($shippingRates as $shippingRateId => $shippingRate) {
            if (!$bigBuyCarrierOption && ZoneFactory::LOCAL_PICKUP === $shippingRate->method_id) {
                $shippingMethods[$shippingRateId] = $shippingRate;
                break;
            }

            if ($bigBuyCarrierOption && ZoneFactory::LOCAL_PICKUP !== $shippingRate->method_id) {
                $shippingMethods[$shippingRateId] = $shippingRate;
            }
        }

        return $shippingMethods;
    }

    /**
     * @param string $label
     * @param \WC_Shipping_Rate $method
     * @return string
     * @throws WooCommerceApiConnectionException
     */
    public function modifyShippingLabel(string $label, \WC_Shipping_Rate $method): ?string
    {
        $methodSanitize = strtok($method->id, ':');

        if($methodSanitize !== ZoneFactory::LOCAL_PICKUP){
            return $label;
        }

        $selectedCarrier = $this->getCarriersResponse();

        if (array_key_exists('carrier_name', $selectedCarrier) && array_key_exists('msg', $selectedCarrier)) {
            return  $selectedCarrier['carrier_name'] . ' ' .$selectedCarrier['msg'];
        }

        if(!$selectedCarrier){
            return ZoneFactory::ALL_WORLD_ZONE;
        }

        if (array_key_exists('carrier_delay', $selectedCarrier)
            && array_key_exists('carrier_cost', $selectedCarrier)
            && array_key_exists('carrier_name', $selectedCarrier)){
            return $selectedCarrier['carrier_name'].' ' .$selectedCarrier['carrier_delay'].' - '
                .str_replace('.', ',', $selectedCarrier['carrier_cost']).'€';
        }

        return $label;
    }

    /**
     * @param \WC_Shipping_Rate $method
     * @throws WooCommerceApiConnectionException
     */
    public function findDataToProcess(\WC_Shipping_Rate $method): void
    {
        if(!is_checkout()) {
            return;
        }

        $methodSanitize = strtok($method->id, ':');

        if($methodSanitize !== ZoneFactory::LOCAL_PICKUP ) {
            return;
        }

        $selectedCarrier = $this->getCarriersResponse();

        if(!$selectedCarrier) {
            return;
        }
    }

    /**
     * @return array
     * @throws WooCommerceApiConnectionException
     */
    public function getCarriersResponse(): array
    {
        $selectedCarrier = [];
        $productQuantityIndexedByProductId = [];
        $productIdRequestIndexedBySku = [];
        $productsCart = WC()->cart->cart_contents;
        foreach ($productsCart as $productCart){
            if ($productCart['variation_id'] === self::VARIATION_ID){
                $productQuantityIndexedByProductId[$productCart['product_id']] = $productCart['quantity'];
                continue;
            }

            $productQuantityIndexedByProductId[$productCart['variation_id']] = $productCart['quantity'];
        }

        $productSkuIndexedByProductId = $this->productManager->findProductIdAndSkuByProductIds(array_keys($productQuantityIndexedByProductId));
        foreach ($productSkuIndexedByProductId as $productId => $productSku){
            if (array_key_exists($productId, $productQuantityIndexedByProductId)){
                $productIdRequestIndexedBySku['products[' . $productSku . ']'] = $productQuantityIndexedByProductId[$productId];
            }
        }

        if (!empty(WC()->shipping()->packages)) {
            $country = WC()->shipping()->packages[0]['destination']['country'];
            $postcode = WC()->shipping()->packages[0]['destination']['postcode'];
            $selectedCarrier = $this->shippingService->getPackageShippingCost($country, $postcode, $productIdRequestIndexedBySku);
        }

        return $selectedCarrier;
    }

    /**
     * @param \WC_Cart $cart
     *
     * @return \WC_Cart
     * @throws WooCommerceApiConnectionException
     */
    public function getTotalCart(\WC_Cart $cart): \WC_Cart
    {
        $method = WC()->session->get( 'chosen_shipping_methods' );
        $methodSanitize = strtok($method[self::METHOD_POSITION], ':');

        if($methodSanitize !== ZoneFactory::LOCAL_PICKUP ){
            return $cart;
        }

        $selectedCarrier = $this->getCarriersResponse();

        if(!$selectedCarrier || array_key_exists('msg', $selectedCarrier)){
            return $cart;
        }

        if (array_key_exists('carrier_delay', $selectedCarrier)
            && array_key_exists('carrier_cost', $selectedCarrier)
            && array_key_exists('taxes', $selectedCarrier)
            && array_key_exists('id', $selectedCarrier)){
            $shippingMethod = $cart->calculate_shipping();
            $shippingMethod[0]->set_label($selectedCarrier['carrier_name'].' '.$selectedCarrier['carrier_delay']);
            $shippingMethod[0]->set_cost($selectedCarrier['carrier_cost'] - $selectedCarrier['taxes']);
            $shippingMethod[0]->set_taxes([$selectedCarrier['id'] => $selectedCarrier['taxes']]);
            $cart->set_shipping_total(($selectedCarrier['carrier_cost'] - $selectedCarrier['taxes']));
            $cart->set_shipping_tax($selectedCarrier['taxes']);
            $cart->set_shipping_taxes([$selectedCarrier['id'] => $selectedCarrier['taxes']]);
            $subTotal = $cart->subtotal;
            $cart->total = $subTotal += $selectedCarrier['carrier_cost'];
            $subTotalTax = $cart->get_subtotal_tax();
            $cart->set_total_tax($subTotalTax + $selectedCarrier['taxes']);

            return $cart;
        }

        $shippingMethod = $cart->calculate_shipping();
        $shippingMethod[0]->set_label($selectedCarrier['carrier_name'].' '.$selectedCarrier['carrier_delay']);
        $shippingMethod[0]->set_cost($selectedCarrier['carrier_cost']);
        $shippingMethod[0]->set_taxes([0 => 0]);
        $subTotal = $cart->subtotal;
        $cart->total = $subTotal += $selectedCarrier['carrier_cost'];
        $subTotalTax = $cart->get_subtotal_tax();
        $cart->set_total_tax($subTotalTax);

        return $cart;
    }

    /**
     * @throws WooCommerceApiConnectionException
     */
    public function createAllWorldZoneIfNotExists(): void
    {
        try {
            $allWorldZoneId = $this->getAllWorldZoneId();
        } catch (NoResultException $exception) {
            $zoneModel = $this->zoneFactory->create();
            try {
                $zones = $this->zoneManager->createZones(json_decode(json_encode($zoneModel), true));
                $allWorldZoneId = $zones['id'];
            } catch (WooCommerceApiConnectionException $exception) {
                $this->logger->error('Creating All World Zone - Zone Model: '.json_encode($zoneModel).' - Error message: '.$exception->getMessage());
            }
        }

        try {
            $this->enableLocalPickup($allWorldZoneId);
        } catch (NoResultException $exception) {
            $methodModel = $this->methodFactory->setLocalPickupMethod();
            try {
                $this->zoneManager->createShippingZones($allWorldZoneId, json_decode(json_encode($methodModel), true));
            } catch (WooCommerceApiConnectionException $exception) {
                $this->logger->error('Creating Zone Method - Method Model: '.json_encode($methodModel).' - Error message: '.$exception->getMessage());
            }
        }
    }

    /**
     * @return int
     * @throws WooCommerceApiConnectionException
     * @throws NoResultException
     */
    private function getAllWorldZoneId(): int
    {
        $findAllZones = $this->zoneManager->findAllZones();
        $zonesIndexedById = array_column($findAllZones, 'id', 'name');

        if (!array_key_exists(ZoneFactory::ALL_WORLD_ZONE, $zonesIndexedById)) {
            throw new NoResultException('No ID found');
        }

        return $zonesIndexedById[ZoneFactory::ALL_WORLD_ZONE];
    }

    /**
     * @param int $zoneId
     * @return int
     * @throws WooCommerceApiConnectionException
     * @throws NoResultException
     */
    private function getLocalPickupId(int $zoneId): int
    {
        $carriers = [];
        $shippingZone = $this->zoneManager->findShippingZoneByZoneId($zoneId);

        foreach ($shippingZone as $shippingMethod) {
            if ($shippingMethod['enabled'] && $shippingMethod['method_id'] === ZoneFactory::LOCAL_PICKUP) {
                $carriers = array_column($shippingZone, 'instance_id', 'method_id');
                break;
            }
        }

        if (!array_key_exists(ZoneFactory::LOCAL_PICKUP, $carriers)) {
            throw new NoResultException('No ID found');
        }

        return $carriers[ZoneFactory::LOCAL_PICKUP];
    }

    /**
     * @return bool
     * @throws WooCommerceApiConnectionException
     */
    public function checkIfAllWorldZoneLocalPickupExists(): bool
    {
        try {
            $allZoneWorldId = $this->getAllWorldZoneId();
        } catch (NoResultException $e) {
            return false;
        }

        try {
            $this->getLocalPickupId($allZoneWorldId);
        } catch (NoResultException $e) {
            return false;
        }

        return true;
    }

    /**
     * @param int $zoneId
     * @return mixed
     * @throws NoResultException
     * @throws WooCommerceApiConnectionException
     */
    private function enableLocalPickup(int $zoneId)
    {
        $shippingZoneMethodsIndexedById = [];
        $shippingZoneDisabledMethodsIndexedById = [];
        $shippingZone = $this->zoneManager->findShippingZoneByZoneId($zoneId);

        foreach ($shippingZone as $shippingMethod) {
            if (!isset($shippingMethod['enabled'], $shippingMethod['method_id']) || $shippingMethod['method_id'] !== ZoneFactory::LOCAL_PICKUP) {
                continue;
            }

            $shippingZoneIndexedById =  array_column($shippingZone, 'instance_id', 'method_id');
            
            if($shippingMethod['enabled']){
                $shippingZoneMethodsIndexedById = $shippingZoneIndexedById;
                break;
            }
            $shippingZoneDisabledMethodsIndexedById = $shippingZoneIndexedById;
        }

        if(array_key_exists(ZoneFactory::LOCAL_PICKUP, $shippingZoneMethodsIndexedById)){
            return $shippingZoneMethodsIndexedById[ZoneFactory::LOCAL_PICKUP];
        }

        if (!array_key_exists(ZoneFactory::LOCAL_PICKUP, $shippingZoneDisabledMethodsIndexedById)) {
            throw new NoResultException('LocalPickup not found at shippingZoneMethods or shippingZoneDisabledMethods');
        }

        try {
            $localPickUp = $shippingZoneDisabledMethodsIndexedById[ZoneFactory::LOCAL_PICKUP];
            $this->zoneManager->enableShippingZoneMethodByZoneIdAndMethodId($zoneId, $localPickUp);
            return $localPickUp;
        } catch (\Exception $exception) {
            $this->logger->error('Enabling shipping zone method id: '.$localPickUp.' - Error message: '.$exception->getMessage());
            throw new NoResultException('Enabling shipping zone method id: '.$localPickUp.' - Error message: '.$exception->getMessage());
        }
    }
}